import functools
import itertools
import warnings
from collections.abc import (
    Collection,
    Iterator,
)
from datetime import datetime
from typing import (
    Any,
    Callable,
    Optional,
    Union,
)

import neptune_scale.api.run

__all__ = ("Run",)

from neptune_scale.sync.metadata_splitter import Metrics


def warn_unsupported_params(fn: Callable) -> Callable:
    # Perform some simple heuristics to detect if a method is called with parameters
    # that are not supported by Scale
    warn = functools.partial(warnings.warn, stacklevel=3)

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):  # type: ignore
        if kwargs.get("wait") is not None:
            warn("The `wait` parameter is not yet implemented and will be ignored.")

        expected = {
            "wait",
            "step",
            "timestamp",
            "steps",
            "timestamps",
            "preview",
            "previews",
            "preview_completion",
            "preview_completions",
        }
        extra_kwargs = set(kwargs.keys()) - expected
        if extra_kwargs:
            warn(
                f"`{fn.__name__}()` was called with additional keyword argument(s): `{', '.join(extra_kwargs)}`. "
                "These arguments are not supported by Neptune and will be ignored."
            )

        return fn(*args, **kwargs)

    return wrapper


class Attribute:
    """Objects of this class are returned on dict-like access to Run. Attributes have a path and
    allow logging values under it. Example:

        run = Run(...)
        run['foo'] = 1
        run['nested'] = {'foo': 1, {'bar': {'baz': 2}}}
        run['bar'].append(1, step=10)
    """

    def __init__(
        self,
        run: "Run",
        path: str,
    ) -> None:
        self._log = run._log
        self._path = path

    @warn_unsupported_params
    def assign(self, value: Any, *, wait: bool = False) -> None:
        self._log(configs=accumulate_dict_values(value, self._path))

    @warn_unsupported_params
    def append(
        self,
        value: Union[dict[str, Any], float],
        *,
        step: Union[float, int],
        preview: bool = False,
        preview_completion: Optional[float] = None,
        timestamp: Optional[Union[float, datetime]] = None,
        wait: bool = False,
        **kwargs: Any,
    ) -> None:
        if isinstance(timestamp, float):
            timestamp = datetime.fromtimestamp(timestamp)

        self._log(
            timestamp=timestamp,
            step=step,
            metrics=Metrics(
                data=accumulate_dict_values(value, self._path),
                preview=preview,
                preview_completion=preview_completion,
            ),
        )

    @warn_unsupported_params
    def add(self, values: Union[str, Union[list[str], set[str], tuple[str]]], *, wait: bool = False) -> None:
        if isinstance(values, str):
            values = (values,)
        self._log(tags_add={self._path: values})

    @warn_unsupported_params
    def remove(self, values: Union[str, Union[list[str], set[str], tuple[str]]], *, wait: bool = False) -> None:
        if isinstance(values, str):
            values = (values,)
        self._log(tags_remove={self._path: values})

    @warn_unsupported_params
    def extend(
        self,
        values: Collection[Union[float, int]],
        *,
        steps: Collection[Union[float, int]],
        timestamps: Optional[Collection[Union[float, datetime]]] = None,
        previews: Optional[Collection[bool]] = None,
        preview_completions: Optional[Collection[float]] = None,
        wait: bool = False,
        **kwargs: Any,
    ) -> None:
        self._validate_lists_length_equal(values, steps, "steps")
        self._validate_lists_length_equal(values, timestamps, "timestamps")
        self._validate_lists_length_equal(values, previews, "preview")
        self._validate_lists_length_equal(values, preview_completions, "preview_completions")
        if timestamps is None:
            timestamps = (datetime.now(),) * len(values)
        it = itertools.zip_longest(values, steps, timestamps, previews or [], preview_completions or [])
        for value, step, ts, preview, completion in it:
            kwargs = {}
            if preview is not None:
                kwargs["preview"] = preview
            if completion is not None:
                kwargs["preview_completion"] = completion
            self.append(value, step=step, timestamp=ts, wait=wait, **kwargs)

    @staticmethod
    def _validate_lists_length_equal(values: Collection[Any], other: Optional[Collection[Any]], name: str) -> None:
        if other is not None:
            if len(other) != len(values):
                raise ValueError(f"All lists provided to extend must have the same length; len(value) != len({name})")


class Run(neptune_scale.api.run.Run):
    """This class extends the main Run class with a dict-like API compatible (on a basic level)
    with the legacy neptune-client package.

    Example:

        from neptune_scale.legacy import Run

        run = Run(...)
        run['foo'] = 1
        run['metrics/loss'].append(0.5, step=10)

        run.close()
    """

    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)
        self._attributes: dict[str, Attribute] = {}

    def __getitem__(self, path: str) -> "Attribute":
        path = cleanup_path(path)
        attr = self._attributes.get(path)
        if attr is None:
            attr = Attribute(self, path)
            self._attributes[path] = attr

        return attr

    def __setitem__(self, key: str, value: Any) -> None:
        # TODO: validate type if attr is already known?
        attr = self[key]
        attr.assign(value)


def cleanup_path(path: str) -> str:
    """
    >>> cleanup_path('/a/b/c')
    'a/b/c'
    >>> cleanup_path('a/b/c/')
    Traceback (most recent call last):
    ...
    ValueError: Invalid path: `a/b/c/`. Path must not end with a slash.
    >>> cleanup_path('a//b/c')
    Traceback (most recent call last):
    ...
    ValueError: Invalid path: `a//b/c`. Path components must not be empty.
    >>> cleanup_path('a/ /b/c')
    Traceback (most recent call last):
    ...
    ValueError: Invalid path: `a/ /b/c`. Path components cannot contain leading or trailing whitespace.
    >>> cleanup_path('a/b/c ')
    Traceback (most recent call last):
    ...
    ValueError: Invalid path: `a/b/c `. Path components cannot contain leading or trailing whitespace.
    """

    if path.strip() in ("", "/"):
        raise ValueError(f"Invalid path: `{path}`.")

    orig_parts = path.split("/")
    parts = [x.strip() for x in orig_parts]

    for i, part in enumerate(parts):
        if part != orig_parts[i]:
            raise ValueError(f"Invalid path: `{path}`. Path components cannot contain leading or trailing whitespace.")

    # Skip the first slash, if present
    if parts[0] == "":
        parts = parts[1:]

    if parts[-1] == "":
        raise ValueError(f"Invalid path: `{path}`. Path must not end with a slash.")

    if not all(parts):
        raise ValueError(f"Invalid path: `{path}`. Path components must not be empty.")

    return "/".join(parts)


def accumulate_dict_values(value: Union[Any, dict[str, Any]], path_or_base: str) -> dict[str, Any]:
    """
    If value is a dict, flatten nested dictionaries into a single dict with unwrapped paths, each
    starting with `path_or_base`.

    If value is an atom, return a dict with a single entry `path_or_base` -> `value`.

    >>> accumulate_dict_values(1, "foo")
    {'foo': 1}
    >>> accumulate_dict_values({"bar": 1, 'l0/l1': 2, 'l3':{"l4": 3}}, "foo")
    {'foo/bar': 1, 'foo/l0/l1': 2, 'foo/l3/l4': 3}
    """

    if isinstance(value, dict):
        data = {"/".join(path): value for path, value in iter_nested(value, path_or_base)}
    else:
        data = {path_or_base: value}

    return data


def iter_nested(dict_: dict[str, Any], path: str) -> Iterator[tuple[tuple[str, ...], Any]]:
    """Iterate a nested dictionary, yielding a tuple of path components and value.

    >>> list(iter_nested({"foo": 1, "bar": {"baz": 2}}, "base"))
    [(('base', 'foo'), 1), (('base', 'bar', 'baz'), 2)]
    >>> list(iter_nested({"foo":{"bar": 1}, "bar":{"baz": 2}}, "base"))
    [(('base', 'foo', 'bar'), 1), (('base', 'bar', 'baz'), 2)]
    >>> list(iter_nested({"foo": 1, "bar": 2}, "base"))
    [(('base', 'foo'), 1), (('base', 'bar'), 2)]
    >>> list(iter_nested({"foo": {}}, ""))
    Traceback (most recent call last):
    ...
    ValueError: The dictionary cannot be empty or contain empty nested dictionaries.
    """

    parts = tuple(path.split("/"))
    yield from _iter_nested(dict_, parts)


def _iter_nested(dict_: dict[str, Any], path_acc: tuple[str, ...]) -> Iterator[tuple[tuple[str, ...], Any]]:
    if not dict_:
        raise ValueError("The dictionary cannot be empty or contain empty nested dictionaries.")

    for key, value in dict_.items():
        current_path = path_acc + (key,)
        if isinstance(value, dict):
            yield from _iter_nested(value, current_path)
        else:
            yield current_path, value



__all__ = [
    "Run",
    "NeptuneLoggingHandler",
]

import warnings

from neptune_scale.api.run import Run
from neptune_scale.exceptions import NeptuneScaleWarning
from neptune_scale.logging.logging_handler import NeptuneLoggingHandler

warnings.simplefilter("once", category=NeptuneScaleWarning)



from pathlib import Path

import azure.core.exceptions
from azure.core.pipeline.transport import AsyncioRequestsTransport
from azure.storage.blob.aio import BlobClient

from neptune_scale.exceptions import NeptuneFileUploadTemporaryError
from neptune_scale.util import get_logger

__all__ = ["upload_to_azure"]

logger = get_logger()


async def upload_to_azure(local_path: str, mime_type: str, storage_url: str, chunk_size: int = 4 * 1024 * 1024) -> None:
    logger.debug(f"Starting upload to Azure: {local_path}, {mime_type=}, {chunk_size=}")

    try:
        size_bytes = Path(local_path).stat().st_size
        with open(local_path, "rb") as file:
            client = BlobClient.from_blob_url(
                storage_url,
                max_block_size=chunk_size,
                max_initial_backoff=5,
                increment_base=3,
                retry_total=5,
                transport=AsyncioRequestsTransport(),
            )
            async with client:
                # Upload with default concurrency settings
                await client.upload_blob(
                    file,
                    content_type=mime_type,
                    overwrite=True,
                    length=size_bytes,
                )
    except azure.core.exceptions.AzureError as e:
        logger.debug(f"Azure SDK error, will retry uploading file {local_path}: {e}")
        raise NeptuneFileUploadTemporaryError() from e
    except Exception as e:
        logger.debug(f"Failed to upload file {local_path}: {e}")
        raise e

    logger.debug(f"Finished upload to Azure: {local_path}")



from __future__ import annotations

import functools
import logging
import pickle
from pathlib import Path

from neptune_scale.sync.parameters import MAX_SINGLE_OPERATION_SIZE_BYTES

__all__ = (
    "OperationsRepository",
    "OperationType",
    "Operation",
    "Metadata",
    "SequenceId",
    "FileUploadRequest",
    "OperationSubmission",
)

import contextlib
import os
import sqlite3
import threading
import time
from collections.abc import (
    Generator,
    Iterable,
)
from contextlib import (
    AbstractContextManager,
    nullcontext,
)
from dataclasses import dataclass
from datetime import datetime
from enum import IntEnum
from typing import (
    Any,
    Callable,
    Literal,
    NewType,
    Optional,
    TypeVar,
    Union,
)

from neptune_api.proto.neptune_pb.ingest.v1.common_pb2 import Run as CreateRun
from neptune_api.proto.neptune_pb.ingest.v1.common_pb2 import UpdateRunSnapshot

from neptune_scale.exceptions import (
    NeptuneLocalStorageInUnsupportedVersion,
    NeptuneScaleError,
    NeptuneScaleWarning,
    NeptuneUnableToLogData,
    NeptuneUnexpectedError,
)
from neptune_scale.util import (
    envs,
    get_logger,
)

logger = get_logger()

DB_VERSION = "v5"
BACKWARD_COMPATIBLE_DB_VERSIONS = ("v5",)

ErrorId = NewType("ErrorId", int)
SequenceId = NewType("SequenceId", int)
RequestId = NewType("RequestId", str)


T = TypeVar("T")
R = TypeVar("R")


def log_timing(func: Callable[..., R]) -> Callable[..., R]:
    """Decorator that logs execution time (in ms) of the decorated callable at DEBUG level on exit.
    If DEBUG is disabled, the original function is executed with near-zero overhead.
    """
    if not logger.isEnabledFor(logging.DEBUG):
        return func

    min_time_to_log_call_ms = float(os.getenv(envs.LOG_TIMING_THRESHOLD_MS, "1000.0"))

    def represent_argument(name: Optional[str], value: object) -> str:
        if isinstance(value, OperationsRepository):
            return "self"

        name_str = f"{name}=" if name is not None else ""
        value_str = repr(value)
        return f"{name_str}{value_str[:100]}..." if len(value_str) > 100 else value_str

    @contextlib.contextmanager
    def timer(caller: str, args: Any = None, kwargs: Any = None) -> Generator[None, None, None]:
        start = time.perf_counter()
        yield
        elapsed_ms = (time.perf_counter() - start) * 1000
        if elapsed_ms >= min_time_to_log_call_ms:
            argument_representation = ", ".join(
                [represent_argument(None, v) for v in (args or [])]
                + [f"{k}={represent_argument(k, v)}" for k, v in (kwargs or {}).items() if k != "timer"]
            )
            pid = os.getpid()
            logger.debug(f"PID={pid} {caller}({argument_representation}) executed in {elapsed_ms:.2f} ms")

    @functools.wraps(func)
    def wrapper(*args: T, **kwargs: T) -> R:
        return func(*args, timer=timer(func.__qualname__, args, kwargs), **kwargs)

    return wrapper


def log_diagnostics(db_path: Path) -> None:
    if logger.isEnabledFor(logging.DEBUG):
        try:
            db_size = db_path.stat().st_size if db_path.exists() else None

            wal_path = db_path.with_suffix(db_path.suffix + "-wal")
            wal_size = wal_path.stat().st_size if wal_path.exists() else None

            shm_path = db_path.with_suffix(db_path.suffix + "-shm")
            shm_size = shm_path.stat().st_size if shm_path.exists() else None

            logger.debug(f"Database diagnostics: db size={db_size}, wal size={wal_size}, shm size={shm_size}")
        except Exception:
            logger.error("Failed to get database diagnostics", exc_info=True)


class OperationType(IntEnum):
    UPDATE_SNAPSHOT = 0
    CREATE_RUN = 1


@dataclass(frozen=True)
class Operation:
    sequence_id: SequenceId
    timestamp: int
    operation_type: OperationType
    operation: Union[UpdateRunSnapshot, CreateRun]
    operation_size_bytes: int

    @property
    def ts(self) -> datetime:
        return datetime.fromtimestamp(self.timestamp / 1000)


@dataclass(frozen=True)
class OperationSubmission:
    sequence_id: SequenceId
    timestamp: int
    request_id: RequestId

    @property
    def ts(self) -> datetime:
        return datetime.fromtimestamp(self.timestamp / 1000)


@dataclass(frozen=True)
class OperationError:
    error_id: Optional[ErrorId]
    timestamp: int
    error_type: str
    error_details: str
    error_body: str
    sequence_id: Optional[SequenceId] = None

    @property
    def ts(self) -> datetime:
        return datetime.fromtimestamp(self.timestamp / 1000)

    @staticmethod
    def serialize_error(error: BaseException, sequence_id: Optional[SequenceId] = None) -> OperationError:
        try:
            if not isinstance(error, (NeptuneScaleError, NeptuneScaleWarning)):
                serialized_error: BaseException = NeptuneUnexpectedError(reason=str(error))
            else:
                serialized_error = error
            error_body = pickle.dumps(serialized_error).hex()
        except (pickle.PicklingError, TypeError):
            logger.error(f"Failed to serialize error {error}", exc_info=True)
            error_body = ""

        return OperationError(
            error_id=None,  # This will be set when saving to the database
            timestamp=int(time.time() * 1000),
            error_type=type(error).__name__,
            error_details=str(error),
            error_body=error_body,
            sequence_id=sequence_id,
        )

    def deserialize_error(self) -> BaseException:
        try:
            result = pickle.loads(bytes.fromhex(self.error_body))
            if isinstance(result, (NeptuneScaleError, NeptuneScaleWarning)):
                return result
            else:
                logger.error(f"Failed to deserialize error {self.error_type} {self.error_details}", exc_info=True)
                return NeptuneUnexpectedError(reason=f"{self.error_type} {self.error_details}")
        except (pickle.UnpicklingError, ValueError):
            logger.error(f"Failed to deserialize error {self.error_type} {self.error_details}", exc_info=True)
            return NeptuneUnexpectedError(reason=f"{self.error_type} {self.error_details}")


@dataclass(frozen=True)
class Metadata:
    project: str
    run_id: str


@dataclass(frozen=True)
class FileUploadRequest:
    source_path: str
    destination: str
    mime_type: str
    size_bytes: int
    is_temporary: bool = False
    sequence_id: Optional[SequenceId] = None


class OperationsRepository:
    """A disk-based repository for Neptune operations using SQLite.

    This repository stores and manages Neptune operations in a SQLite database with the following schema:
    - Table `run_operations`: Contains serialized RunOperation objects
      - sequence_id INTEGER: Auto-incremented operation ID (Primary Key)
      - timestamp INTEGER: Timestamp of the operation being enqueued
      - operation_type INTEGER NOT NULL: Type of operation (0=snapshot, 1=create_run)
      - operation BLOB NOT NULL: Serialized RunOperation object (protobuf)
      - operation_size_bytes INTEGER NOT NULL: Size of the operation

    - Table `metadata`: Contains metadata about runs
      - id INTEGER: Auto-incremented ID (Primary Key)
      - version TEXT: Version identifier
      - project TEXT: Project identifier
      - run_id TEXT: Run identifier
    """

    def __init__(
        self,
        db_path: Path,
        timeout: Optional[float] = None,
    ) -> None:
        if not db_path.is_absolute():
            raise RuntimeError("db_path must be an absolute path")

        self._db_path = db_path
        self._lock = threading.RLock()
        self._connection: Optional[sqlite3.Connection] = None

        self._timeout = (
            timeout
            if timeout is not None
            else envs.get_positive_int(envs.OPERATION_REPOSITORY_TIMEOUT_MS, 60_000) / 1000.0
        )

        self._log_failure_action: Literal["raise", "drop"] = envs.get_option(  # type: ignore
            envs.LOG_FAILURE_ACTION, ("drop", "raise"), "drop"
        )

    @log_timing
    def init_db(self, timer: Optional[AbstractContextManager] = None) -> None:
        logger.debug(f"Create SQLite database file {self._db_path}")
        os.makedirs(self._db_path.parent, exist_ok=True)
        with self._get_connection(timer=timer) as conn:  # type: ignore
            conn.execute(
                """
                CREATE TABLE IF NOT EXISTS run_operations (
                    sequence_id INTEGER PRIMARY KEY AUTOINCREMENT,
                    timestamp INTEGER NOT NULL,
                    operation_type INTEGER NOT NULL,
                    operation BLOB NOT NULL,
                    operation_size_bytes INTEGER NOT NULL
                )
                """
            )

            # This index is necessary for the get_operations method to work efficiently
            # It allows the initial query (retrieving sizes) to be done without reading the operations.
            conn.execute(
                """
                CREATE INDEX IF NOT EXISTS idx_run_operations_sequence_size
                    ON run_operations (sequence_id, operation_size_bytes);
                """
            )

            conn.execute(
                """
                CREATE TABLE IF NOT EXISTS metadata (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    version TEXT NOT NULL,
                    project TEXT NOT NULL,
                    run_id TEXT NOT NULL
                )"""
            )

            conn.execute(
                """
                CREATE TABLE IF NOT EXISTS run_operation_submission (
                    sequence_id INTEGER PRIMARY KEY,
                    timestamp INTEGER NOT NULL,
                    request_id TEXT NOT NULL
                )
                """
            )

            conn.execute(
                """
                CREATE TABLE IF NOT EXISTS file_upload_requests (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    source_path TEXT NOT NULL,
                    destination TEXT NOT NULL,
                    mime_type TEXT NOT NULL,
                    size_bytes INTEGER NOT NULL,
                    is_temporary INTEGER NOT NULL
                )
                """
            )

            conn.execute(
                """
                CREATE TABLE IF NOT EXISTS operation_errors (
                    error_id INTEGER PRIMARY KEY AUTOINCREMENT,
                    timestamp INTEGER NOT NULL,
                    error_type TEXT NOT NULL,
                    error_details TEXT NOT NULL,
                    error_body TEXT NOT NULL,
                    sequence_id INTEGER
                )
                """
            )

    @log_timing
    def save_update_run_snapshots(
        self, updates: list[UpdateRunSnapshot], timer: Optional[AbstractContextManager] = None
    ) -> SequenceId:
        """
        Guarantees:
        - The order of operations saved to the DB is preserved within a single call.
        - The timestamp is consistent for all operations in a single call.
        - When called by multiple threads, operations are not interleaved between threads.

        Returns -1 when the list of operations is empty.

        This method saves each chunk of operations in a separate transaction, so that
        a single large batch doesn’t lock the database for an extended period of time.
        """

        if not updates:
            return SequenceId(-1)

        # Chunk the operations so that no batch exceeds MAX_SINGLE_OPERATION_SIZE_BYTES.
        chunked_ops = self._chunk_operations(updates)

        last_insert_rowid = SequenceId(-1)
        with self._lock:
            current_time = int(time.time() * 1000)  # millisecond timestamp
            for chunk in chunked_ops:
                insert_rowid = self._insert_update_run_snapshots(chunk, current_time)
                if insert_rowid is not None:
                    last_insert_rowid = insert_rowid

        return last_insert_rowid

    @staticmethod
    def _chunk_operations(
        updates: list[UpdateRunSnapshot], timer: Optional[AbstractContextManager] = None
    ) -> list[list[bytes]]:
        """
        Split operations into batches so that each batch does not exceed MAX_SINGLE_OPERATION_SIZE_BYTES.
        Returns serialized operations in a list of batches.
        """
        batches: list[list[bytes]] = []
        current_batch: list[bytes] = []
        current_batch_size = 0

        for update in updates:
            serialized = update.SerializeToString()
            size_bytes = len(serialized)

            if size_bytes > MAX_SINGLE_OPERATION_SIZE_BYTES:
                raise RuntimeError(
                    f"Operation size ({size_bytes}) exceeds the limit of " f"{MAX_SINGLE_OPERATION_SIZE_BYTES} bytes"
                )

            # If adding this operation would exceed the limit, start a new batch.
            if current_batch_size + size_bytes > MAX_SINGLE_OPERATION_SIZE_BYTES:
                batches.append(current_batch)
                current_batch = []
                current_batch_size = 0

            current_batch.append(serialized)
            current_batch_size += size_bytes

        # Add the last batch if it's not empty
        if current_batch:
            batches.append(current_batch)

        return batches

    @log_timing
    def _insert_update_run_snapshots(
        self, ops: list[bytes], current_time: int, timer: Optional[AbstractContextManager] = None
    ) -> Optional[SequenceId]:
        """
        Inserts operations into 'run_operations' in a single transaction.
        """
        try:
            with self._get_connection(timer=timer) as conn:  # type: ignore
                with contextlib.closing(conn.cursor()) as cursor:
                    cursor.executemany(
                        """
                        INSERT INTO run_operations (timestamp, operation_type, operation, operation_size_bytes)
                        VALUES (?, ?, ?, ?)
                        """,
                        [
                            (current_time, OperationType.UPDATE_SNAPSHOT, serialized_op, len(serialized_op))
                            for serialized_op in ops
                        ],
                    )
                    cursor.execute("SELECT last_insert_rowid()")
                    return SequenceId(cursor.fetchone()[0])

        except NeptuneUnableToLogData:
            if self._log_failure_action == "raise":
                raise
            else:
                logger.error(f"Dropping {len(ops)} operations due to error", exc_info=True)
                return None

    @log_timing
    def save_create_run(self, run: CreateRun, timer: Optional[AbstractContextManager] = None) -> SequenceId:
        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                current_time = int(time.time() * 1000)  # milliseconds timestamp
                serialized_operation = run.SerializeToString()
                operation_size_bytes = len(serialized_operation)

                cursor.execute(
                    "INSERT INTO run_operations (timestamp, operation_type, operation, operation_size_bytes) VALUES (?, ?, ?, ?)",
                    (current_time, OperationType.CREATE_RUN, serialized_operation, operation_size_bytes),
                )
                return SequenceId(cursor.lastrowid)

    @log_timing
    def get_operations(
        self,
        up_to_bytes: int,
        from_exclusive: Optional[SequenceId] = None,
        timer: Optional[AbstractContextManager] = None,
    ) -> list[Operation]:
        if up_to_bytes < MAX_SINGLE_OPERATION_SIZE_BYTES:
            raise RuntimeError(
                f"up to bytes is too small: {up_to_bytes} bytes, minimum is {MAX_SINGLE_OPERATION_SIZE_BYTES} bytes"
            )

        if from_exclusive is None:
            from_exclusive = SequenceId(-1)

        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                cursor.execute(
                    """
                    WITH running_size AS (
                        SELECT
                            sequence_id,
                            SUM(operation_size_bytes) FILTER (WHERE sequence_id > ?) OVER (ORDER BY sequence_id ASC) AS cumulative_size
                        FROM run_operations
                        ORDER BY sequence_id ASC
                    )
                    SELECT sequence_id, timestamp, operation, operation_type, operation_size_bytes
                    FROM run_operations
                    WHERE sequence_id > ? AND sequence_id < IFNULL((
                        SELECT sequence_id
                        FROM running_size
                        WHERE cumulative_size > ?
                        LIMIT 1
                    ), (
                        SELECT MAX(sequence_id) + 1 FROM run_operations
                    ))
                    ORDER BY sequence_id ASC
                    """,
                    (
                        from_exclusive,
                        from_exclusive,
                        up_to_bytes,
                    ),
                )

                rows = cursor.fetchall()

        return [_deserialize_operation(row) for row in rows]

    @log_timing
    def delete_operations(self, up_to_seq_id: SequenceId, timer: Optional[AbstractContextManager] = None) -> int:
        if up_to_seq_id <= 0:
            return 0

        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                cursor.execute("DELETE FROM run_operations WHERE sequence_id <= ?", (up_to_seq_id,))

                # Return the number of rows affected
                return cursor.rowcount or 0

    @log_timing
    def get_operation_count(self, limit: Optional[int] = None, timer: Optional[AbstractContextManager] = None) -> int:
        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                return self._get_table_count(cursor, "run_operations", limit=limit)

    @log_timing
    def get_operations_sequence_id_range(
        self, timer: Optional[AbstractContextManager] = None
    ) -> Optional[tuple[SequenceId, SequenceId]]:
        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                cursor.execute(
                    """
                    SELECT MIN(sequence_id), MAX(sequence_id)
                    FROM run_operations
                    """
                )

                row = cursor.fetchone()
                if not row:
                    return None

                min_seq_id, max_seq_id = row
                if min_seq_id is None or max_seq_id is None:
                    return None
                return SequenceId(min_seq_id), SequenceId(max_seq_id)

    @log_timing
    def get_operations_min_timestamp(self, timer: Optional[AbstractContextManager] = None) -> Optional[datetime]:
        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                cursor.execute(
                    """
                    SELECT timestamp
                    FROM run_operations
                    ORDER BY sequence_id ASC
                    LIMIT 1
                    """
                )

                row = cursor.fetchone()
                if not row:
                    return None

                (timestamp,) = row
                return datetime.fromtimestamp(timestamp / 1000)

    @log_timing
    def save_metadata(self, project: str, run_id: str, timer: Optional[AbstractContextManager] = None) -> None:
        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                # Check if metadata already exists
                cursor.execute(
                    """
                    SELECT COUNT(*) FROM metadata
                    """
                )

                count = cursor.fetchone()[0]
                if count > 0:
                    raise RuntimeError("Metadata already exists")

                # Insert new metadata
                cursor.execute(
                    """
                    INSERT INTO metadata (version, project, run_id)
                    VALUES (?, ?, ?)
                    """,
                    (DB_VERSION, project, run_id),
                )

    @log_timing
    def get_metadata(self, timer: Optional[AbstractContextManager] = None) -> Optional[Metadata]:
        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                cursor.execute(
                    """
                    SELECT version, project, run_id
                    FROM metadata
                    """
                )

                row = cursor.fetchone()
                if not row:
                    return None

                version, project, run_id = row

                if version not in BACKWARD_COMPATIBLE_DB_VERSIONS:
                    raise NeptuneLocalStorageInUnsupportedVersion()

                return Metadata(project=project, run_id=run_id)

    @log_timing
    def save_file_upload_requests(
        self, files: list[FileUploadRequest], timer: Optional[AbstractContextManager] = None
    ) -> SequenceId:
        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                cursor.executemany(
                    """
                    INSERT INTO file_upload_requests (source_path, destination, mime_type, size_bytes, is_temporary)
                    VALUES (?, ?, ?, ?, ?)
                    """,
                    [
                        (file.source_path, file.destination, file.mime_type, file.size_bytes, int(file.is_temporary))
                        for file in files
                    ],
                )
                cursor.execute("SELECT last_insert_rowid()")
                return SequenceId(cursor.fetchone()[0])

    @log_timing
    def get_file_upload_requests(
        self, n: int, timer: Optional[AbstractContextManager] = None
    ) -> list[FileUploadRequest]:
        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                cursor.execute(
                    """
                    SELECT id, source_path, destination, mime_type, size_bytes, is_temporary
                    FROM file_upload_requests
                    LIMIT ?
                    """,
                    (n,),
                )

                rows = cursor.fetchall()

        return [
            FileUploadRequest(
                sequence_id=SequenceId(row[0]),
                source_path=row[1],
                destination=row[2],
                mime_type=row[3],
                size_bytes=row[4],
                is_temporary=bool(row[5]),
            )
            for row in rows
        ]

    @log_timing
    def delete_file_upload_requests(
        self, seq_ids: list[SequenceId], timer: Optional[AbstractContextManager] = None
    ) -> None:
        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                cursor.executemany(
                    """
                    DELETE FROM file_upload_requests
                    WHERE id = ?
                    """,
                    [(seq_id,) for seq_id in seq_ids],
                )

    @log_timing
    def get_file_upload_requests_count(
        self, limit: Optional[int] = None, timer: Optional[AbstractContextManager] = None
    ) -> int:
        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                return self._get_table_count(cursor, "file_upload_requests", limit=limit)

    @log_timing
    def save_operation_submissions(
        self, submissions: list[OperationSubmission], timer: Optional[AbstractContextManager] = None
    ) -> SequenceId:
        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                cursor.executemany(
                    """
                    INSERT INTO run_operation_submission (sequence_id, timestamp, request_id)
                    VALUES (?, ?, ?)
                    """,
                    [(status.sequence_id, status.timestamp, status.request_id) for status in submissions],
                )
                cursor.execute("SELECT last_insert_rowid()")
                return SequenceId(cursor.fetchone()[0])

    @log_timing
    def get_operation_submissions(
        self, limit: int, timer: Optional[AbstractContextManager] = None
    ) -> list[OperationSubmission]:
        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                cursor.execute(
                    """
                    SELECT sequence_id, timestamp, request_id
                    FROM run_operation_submission
                    ORDER BY sequence_id ASC
                    LIMIT ?
                    """,
                    (limit,),
                )

                rows = cursor.fetchall()
        return [
            OperationSubmission(
                sequence_id=SequenceId(row[0]),
                timestamp=row[1],
                request_id=RequestId(row[2]),
            )
            for row in rows
        ]

    @log_timing
    def delete_operation_submissions(
        self, up_to_seq_id: Optional[SequenceId], timer: Optional[AbstractContextManager] = None
    ) -> int:
        if up_to_seq_id is not None and up_to_seq_id <= 0:
            return 0

        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                if up_to_seq_id is None:
                    cursor.execute("DELETE FROM run_operation_submission")
                else:
                    cursor.execute(
                        "DELETE FROM run_operation_submission WHERE sequence_id <= ?",
                        (up_to_seq_id,),
                    )

                return cursor.rowcount or 0

    @log_timing
    def get_operation_submission_count(
        self, limit: Optional[int] = None, timer: Optional[AbstractContextManager] = None
    ) -> int:
        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                return self._get_table_count(cursor, "run_operation_submission", limit=limit)

    @log_timing
    def get_operation_submission_sequence_id_range(
        self, timer: Optional[AbstractContextManager] = None
    ) -> Optional[tuple[SequenceId, SequenceId]]:
        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                cursor.execute(
                    """
                    SELECT MIN(sequence_id), MAX(sequence_id)
                    FROM run_operation_submission
                    """
                )

                row = cursor.fetchone()
                if not row:
                    return None

                min_seq_id, max_seq_id = row
                if min_seq_id is None or max_seq_id is None:
                    return None
                return SequenceId(min_seq_id), SequenceId(max_seq_id)

    @log_timing
    def get_errors(self, limit: int, timer: Optional[AbstractContextManager] = None) -> list[OperationError]:
        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                cursor.execute(
                    """
                    SELECT error_id, timestamp, error_type, error_details,  error_body, sequence_id
                    FROM operation_errors
                    ORDER BY timestamp ASC
                    LIMIT ?
                    """,
                    (limit,),
                )

                rows = cursor.fetchall()

        return [
            OperationError(
                error_id=ErrorId(row[0]) if row[0] is not None else None,
                timestamp=row[1],
                error_type=row[2],
                error_details=row[3],
                error_body=row[4],
                sequence_id=SequenceId(row[5]) if row[5] is not None else None,
            )
            for row in rows
        ]

    @log_timing
    def save_errors(
        self,
        errors: Iterable[BaseException],
        sequence_id: Optional[SequenceId] = None,
        timer: Optional[AbstractContextManager] = None,
    ) -> ErrorId:
        operation_errors = [OperationError.serialize_error(error, sequence_id) for error in errors]

        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                cursor.executemany(
                    """
                    INSERT INTO operation_errors (timestamp, error_type, error_details, error_body, sequence_id)
                    VALUES (?, ?, ?, ?, ?)
                    """,
                    [
                        (error.timestamp, error.error_type, error.error_details, error.error_body, error.sequence_id)
                        for error in operation_errors
                    ],
                )
                cursor.execute("SELECT last_insert_rowid()")
                return ErrorId(cursor.fetchone()[0])

    @log_timing
    def delete_errors(self, error_ids: Iterable[ErrorId], timer: Optional[AbstractContextManager] = None) -> None:
        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                cursor.executemany(
                    """
                    DELETE FROM operation_errors
                    WHERE error_id = ?
                    """,
                    [(error_id,) for error_id in error_ids],
                )

    @log_timing
    def delete_all_errors(self, timer: Optional[AbstractContextManager] = None) -> None:
        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                cursor.execute("DELETE FROM operation_errors")

    @log_timing
    def get_errors_count(self, limit: Optional[int] = None, timer: Optional[AbstractContextManager] = None) -> int:
        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                return self._get_table_count(cursor, "operation_errors", limit=limit)

    @log_timing
    def _is_repository_empty(self, timer: Optional[AbstractContextManager] = None) -> bool:
        with self._get_connection(timer=timer) as conn:  # type: ignore
            with contextlib.closing(conn.cursor()) as cursor:
                if self._get_table_count(cursor, "run_operations", limit=1) != 0:
                    return False
                if self._get_table_count(cursor, "file_upload_requests", limit=1) != 0:
                    return False
                return True

    @staticmethod
    @log_timing
    def _get_table_count(
        cursor: sqlite3.Cursor,
        table_name: str,
        limit: Optional[int] = None,
        timer: Optional[AbstractContextManager] = None,
    ) -> int:
        try:
            if limit is None:
                source = table_name
            else:
                source = f"(SELECT * FROM {table_name} LIMIT {limit})"

            cursor.execute(
                f"""
                SELECT COUNT(*)
                FROM {source}
                """
            )

            count = cursor.fetchone()[0]
            return int(count)
        except sqlite3.OperationalError:
            cursor.execute(f"SELECT COUNT(*) FROM sqlite_master WHERE type='table' AND name='{table_name}'")
            count = cursor.fetchone()[0]
            if count == 0:
                return 0
            else:
                raise

    @log_timing
    def close(self, cleanup_files: bool, timer: Optional[AbstractContextManager] = None) -> None:
        with self._lock:
            with timer or nullcontext():
                if self._connection is not None:
                    if cleanup_files and self._is_repository_empty():
                        self._connection.close()
                        self._connection = None
                        try:
                            os.remove(self._db_path)
                        except OSError:
                            logger.debug(f"Failed to delete SQLite database file {self._db_path}", exc_info=True)
                        logger.debug(f"Deleted SQLite database file {self._db_path}")
                    else:
                        self._connection.close()
                        self._connection = None
                        logger.debug(f"Closed SQLite connection for {self._db_path}")

    @contextlib.contextmanager  # type: ignore
    def _get_connection(self, timer: Optional[AbstractContextManager]) -> AbstractContextManager[sqlite3.Connection]:  # type: ignore
        with self._lock:
            with timer or nullcontext():
                if self._connection is None:
                    try:
                        self._connection = sqlite3.connect(
                            self._db_path,
                            timeout=self._timeout,
                            check_same_thread=False,  # we use RLock to synchronize access
                        )

                        self._connection.execute("PRAGMA journal_mode = WAL")
                        self._connection.execute("PRAGMA synchronous = FULL")

                        logger.debug(f"Created new SQLite connection for {self._db_path}")
                    except sqlite3.DatabaseError as e:
                        try:
                            if self._connection:
                                self._connection.close()
                                self._connection = None
                        except sqlite3.DatabaseError:
                            logger.debug(f"Failed to close SQLite connection for {self._db_path}", exc_info=True)
                        log_diagnostics(self._db_path)
                        raise NeptuneUnableToLogData() from e

                self._connection.execute("BEGIN")
                try:
                    yield self._connection
                    self._connection.commit()
                except sqlite3.DatabaseError as e:
                    self._connection.rollback()
                    log_diagnostics(db_path=self._db_path)
                    raise NeptuneUnableToLogData() from e
                except Exception:
                    self._connection.rollback()
                    raise


def _deserialize_operation(row: tuple[int, int, bytes, int, int]) -> Operation:
    sequence_id, timestamp, operation, operation_type, operation_size_bytes = row
    op_type = OperationType(operation_type)

    deserialized_op = (
        UpdateRunSnapshot.FromString(operation)
        if op_type == OperationType.UPDATE_SNAPSHOT
        else CreateRun.FromString(operation)
    )
    return Operation(SequenceId(sequence_id), timestamp, op_type, deserialized_op, operation_size_bytes)



from __future__ import annotations

__all__ = ("ErrorsMonitor",)

import time
from collections.abc import Callable
from typing import Optional

from neptune_scale.exceptions import (
    NeptuneConnectionLostError,
    NeptuneRetryableError,
    NeptuneScaleError,
    NeptuneScaleWarning,
    NeptuneUnexpectedError,
)
from neptune_scale.sync.operations_repository import OperationsRepository
from neptune_scale.sync.parameters import ERRORS_MONITOR_THREAD_SLEEP_TIME
from neptune_scale.util import get_logger
from neptune_scale.util.daemon import Daemon

logger = get_logger()


def default_error_callback(error: BaseException, last_seen_at: Optional[float]) -> None:
    logger.error(error)


def default_network_error_callback(error: BaseException, last_seen_at: Optional[float]) -> None:
    if last_seen_at is None or time.time() - last_seen_at > 5:
        logger.warning(f"A network error occurred: {error}. Retrying...")


def default_warning_callback(error: BaseException, last_seen_at: Optional[float]) -> None:
    logger.warning(error)


class ErrorsMonitor(Daemon):
    def __init__(
        self,
        operations_repository: OperationsRepository,
        on_network_error_callback: Optional[Callable[[BaseException, Optional[float]], None]] = None,
        on_error_callback: Optional[Callable[[BaseException, Optional[float]], None]] = None,
        on_warning_callback: Optional[Callable[[BaseException, Optional[float]], None]] = None,
    ):
        super().__init__(name="ErrorsMonitor", sleep_time=ERRORS_MONITOR_THREAD_SLEEP_TIME)

        self._operations_repository: OperationsRepository = operations_repository
        self._on_network_error_callback: Callable[[BaseException, Optional[float]], None] = (
            on_network_error_callback or default_network_error_callback
        )
        self._on_error_callback: Callable[[BaseException, Optional[float]], None] = (
            on_error_callback or default_error_callback
        )
        self._on_warning_callback: Callable[[BaseException, Optional[float]], None] = (
            on_warning_callback or default_warning_callback
        )
        self._last_raised_timestamps: dict[str, float] = {}

        self._drain_errors()

    def _drain_errors(self) -> None:
        errors_count = self._operations_repository.get_errors_count()
        if errors_count > 0:
            logger.warning(f"Found {errors_count} unprocessed errors from a previous run. Deleting them now.")
            self._operations_repository.delete_all_errors()

    def work(self) -> None:
        while op_errors := self._operations_repository.get_errors(limit=1):
            op_error = op_errors[0]
            last_raised_at = self._last_raised_timestamps.get(op_error.error_type, None)
            self._last_raised_timestamps[op_error.error_type] = time.time()
            error = op_error.deserialize_error()

            try:
                if isinstance(error, NeptuneConnectionLostError):
                    self._on_network_error_callback(error, last_raised_at)
                elif isinstance(error, NeptuneScaleWarning):
                    self._on_warning_callback(error, last_raised_at)
                elif isinstance(error, NeptuneRetryableError):
                    self._on_warning_callback(error, last_raised_at)
                elif isinstance(error, NeptuneScaleError):
                    self._on_error_callback(error, last_raised_at)
                else:
                    self._on_error_callback(NeptuneUnexpectedError(reason=str(error)), last_raised_at)
            except Exception as e:
                # Don't let user errors kill the process
                logger.error(f"An exception occurred in user callback function: {e}")
            finally:
                if op_error.error_id:
                    self._operations_repository.delete_errors([op_error.error_id])



from __future__ import annotations

import asyncio
from dataclasses import dataclass
from pathlib import Path

from neptune_api.models import Provider

from neptune_scale.sync.operations_repository import (
    FileUploadRequest,
    Metadata,
    Operation,
    OperationsRepository,
    OperationSubmission,
    OperationType,
    SequenceId,
)
from neptune_scale.sync.size_util import proto_encoded_bytes_field_size
from neptune_scale.sync.storage.azure_storage import upload_to_azure
from neptune_scale.sync.storage.gcs import upload_to_gcp
from neptune_scale.sync.storage.s3 import (
    upload_to_s3_multipart,
    upload_to_s3_single,
)

__all__ = ("run_sync_process",)

import datetime
import functools as ft
import os
import signal
import threading
from collections.abc import Mapping
from types import FrameType
from typing import (
    Optional,
    TypeVar,
)

import backoff
import psutil
from neptune_api.proto.google_rpc.code_pb2 import Code
from neptune_api.proto.neptune_pb.ingest.v1.common_pb2 import UpdateRunSnapshots
from neptune_api.proto.neptune_pb.ingest.v1.ingest_pb2 import IngestCode
from neptune_api.proto.neptune_pb.ingest.v1.pub.client_pb2 import (
    BulkRequestStatus,
    SubmitResponse,
)
from neptune_api.proto.neptune_pb.ingest.v1.pub.ingest_pb2 import RunOperation

from neptune_scale.exceptions import (
    GenericFloatValueNanInfUnsupported,
    NeptuneAttributePathEmpty,
    NeptuneAttributePathExceedsSizeLimit,
    NeptuneAttributePathInvalid,
    NeptuneAttributePathNonWritable,
    NeptuneAttributeTypeMismatch,
    NeptuneAttributeTypeUnsupported,
    NeptuneConnectionLostError,
    NeptuneFileMetadataExceedsSizeLimit,
    NeptuneFileUploadError,
    NeptuneHistogramBinEdgesContainNaN,
    NeptuneHistogramBinEdgesNotIncreasing,
    NeptuneHistogramTooManyBins,
    NeptuneHistogramValuesLengthMismatch,
    NeptuneInternalServerError,
    NeptunePreviewStepNotAfterLastCommittedStep,
    NeptuneProjectError,
    NeptuneProjectInvalidName,
    NeptuneProjectNotFound,
    NeptuneRetryableError,
    NeptuneRunConflicting,
    NeptuneRunDuplicate,
    NeptuneRunError,
    NeptuneRunForkParentNotFound,
    NeptuneRunInvalidCreationParameters,
    NeptuneRunNotFound,
    NeptuneSeriesPointDuplicate,
    NeptuneSeriesStepNonIncreasing,
    NeptuneSeriesStepNotAfterForkPoint,
    NeptuneSeriesStepTooLarge,
    NeptuneSeriesTimestampDecreasing,
    NeptuneStringSetExceedsSizeLimit,
    NeptuneStringValueExceedsSizeLimit,
    NeptuneSynchronizationStopped,
    NeptuneTooManyRequestsResponseError,
    NeptuneUnableToLogData,
    NeptuneUnauthorizedError,
    NeptuneUnexpectedError,
    NeptuneUnexpectedResponseError,
)
from neptune_scale.net.api_client import (
    ApiClient,
    FileSignRequest,
    with_api_errors_handling,
)
from neptune_scale.sync.parameters import (
    HTTP_REQUEST_MAX_TIME_SECONDS,
    MAX_REQUEST_SIZE_BYTES,
    MAX_REQUESTS_STATUS_BATCH_SIZE,
    SHUTDOWN_TIMEOUT,
    STATUS_TRACKING_THREAD_SLEEP_TIME,
    SYNC_PROCESS_SLEEP_TIME,
    SYNC_THREAD_SLEEP_TIME,
)
from neptune_scale.util import (
    Daemon,
    envs,
    get_logger,
)

T = TypeVar("T")

logger = get_logger()

CODE_TO_ERROR: dict[IngestCode.ValueType, Optional[type[Exception]]] = {
    IngestCode.PROJECT_NOT_FOUND: NeptuneProjectNotFound,
    IngestCode.PROJECT_INVALID_NAME: NeptuneProjectInvalidName,
    IngestCode.RUN_NOT_FOUND: NeptuneRunNotFound,
    IngestCode.RUN_DUPLICATE: NeptuneRunDuplicate,
    IngestCode.RUN_CONFLICTING: NeptuneRunConflicting,
    IngestCode.RUN_FORK_PARENT_NOT_FOUND: NeptuneRunForkParentNotFound,
    IngestCode.RUN_INVALID_CREATION_PARAMETERS: NeptuneRunInvalidCreationParameters,
    IngestCode.FIELD_PATH_EXCEEDS_SIZE_LIMIT: NeptuneAttributePathExceedsSizeLimit,
    IngestCode.FIELD_PATH_EMPTY: NeptuneAttributePathEmpty,
    IngestCode.FIELD_PATH_INVALID: NeptuneAttributePathInvalid,
    IngestCode.FIELD_PATH_NON_WRITABLE: NeptuneAttributePathNonWritable,
    IngestCode.FIELD_TYPE_UNSUPPORTED: NeptuneAttributeTypeUnsupported,
    IngestCode.FIELD_TYPE_CONFLICTING: NeptuneAttributeTypeMismatch,
    IngestCode.SERIES_POINT_DUPLICATE: NeptuneSeriesPointDuplicate,
    IngestCode.SERIES_STEP_NON_INCREASING: NeptuneSeriesStepNonIncreasing,
    IngestCode.SERIES_STEP_NOT_AFTER_FORK_POINT: NeptuneSeriesStepNotAfterForkPoint,
    IngestCode.SERIES_STEP_TOO_LARGE: NeptuneSeriesStepTooLarge,
    IngestCode.SERIES_TIMESTAMP_DECREASING: NeptuneSeriesTimestampDecreasing,
    IngestCode.FLOAT_VALUE_NAN_INF_UNSUPPORTED: GenericFloatValueNanInfUnsupported,
    IngestCode.STRING_VALUE_EXCEEDS_SIZE_LIMIT: NeptuneStringValueExceedsSizeLimit,
    IngestCode.STRING_SET_EXCEEDS_SIZE_LIMIT: NeptuneStringSetExceedsSizeLimit,
    IngestCode.SERIES_PREVIEW_STEP_NOT_AFTER_LAST_COMMITTED_STEP: NeptunePreviewStepNotAfterLastCommittedStep,
    IngestCode.FILE_REF_EXCEEDS_SIZE_LIMIT: NeptuneFileMetadataExceedsSizeLimit,
    IngestCode.HISTOGRAM_BIN_EDGES_CONTAINS_NAN: NeptuneHistogramBinEdgesContainNaN,
    IngestCode.HISTOGRAM_TOO_MANY_BINS: NeptuneHistogramTooManyBins,
    IngestCode.HISTOGRAM_BIN_EDGES_NOT_INCREASING: NeptuneHistogramBinEdgesNotIncreasing,
    IngestCode.HISTOGRAM_VALUES_LENGTH_DOESNT_MATCH_BINS: NeptuneHistogramValuesLengthMismatch,
}


def code_to_exception(code: IngestCode.ValueType) -> Exception:
    if code in CODE_TO_ERROR:
        error = CODE_TO_ERROR[code]
        return error()  # type: ignore
    return NeptuneUnexpectedError(reason=f"Unexpected ingestion error code: {code}")


def run_sync_process(
    operations_repository_path: Path,
    api_token: str,
    project: str,
    family: str,
) -> None:
    logger.info("Data synchronization started")
    stop_event = threading.Event()
    signal.signal(signal.SIGTERM, ft.partial(_handle_stop_signal, stop_event))

    operations_repository = OperationsRepository(db_path=operations_repository_path)
    operations_repository.delete_operation_submissions(up_to_seq_id=None)

    sender_thread = SenderThread(
        api_token=api_token,
        operations_repository=operations_repository,
        family=family,
    )
    status_thread = StatusTrackingThread(
        api_token=api_token,
        project=project,
        operations_repository=operations_repository,
    )
    file_uploader_thread = FileUploaderThread(
        api_token=api_token,
        project=project,
        operations_repository=operations_repository,
    )
    threads = [sender_thread, status_thread, file_uploader_thread]
    parent_process = psutil.Process(os.getpid()).parent()

    def metadata_sending_threads_died() -> bool:
        return not sender_thread.is_alive() or not status_thread.is_alive()

    def interrupt_metadata_sending_threads() -> None:
        if sender_thread.is_alive():
            sender_thread.interrupt()

        # we're interrupting the status thread only if
        # - it's alive, and
        # - the operation statuses in operations repository are empty, and
        # - we know it will stay empty (since sender_thread is dead)
        elif status_thread.is_alive() and operations_repository.get_operation_submission_count(limit=1) == 0:
            status_thread.interrupt()

    def file_uploader_thread_died() -> bool:
        return not file_uploader_thread.is_alive()

    def interrupt_file_uploader_thread() -> None:
        file_uploader_thread.interrupt()

    def all_threads_died() -> bool:
        return all(not t.is_alive() for t in threads)

    def close_all_threads() -> None:
        for t in threads:
            t.interrupt()
        for t in threads:
            t.join(timeout=SHUTDOWN_TIMEOUT)
            t.close()

    for thread in threads:
        thread.start()

    try:
        while not stop_event.is_set():
            for thread in threads:
                thread.join(timeout=SYNC_PROCESS_SLEEP_TIME)

            if metadata_sending_threads_died():
                interrupt_file_uploader_thread()
                interrupt_metadata_sending_threads()

            if file_uploader_thread_died():
                interrupt_metadata_sending_threads()

            if all_threads_died():
                break

            if not _is_process_running(parent_process):
                logger.error("SyncProcess: parent process closed unexpectedly. Exiting")
                break

    except KeyboardInterrupt:
        logger.debug("KeyboardInterrupt received")
    finally:
        logger.info("Data synchronization stopping")
        close_all_threads()
        operations_repository.close(cleanup_files=False)
    logger.info("Data synchronization finished")


def _is_process_running(process: Optional[psutil.Process]) -> bool:
    try:
        # Check if parent exists and is running
        return process is not None and process.is_running()
    except (psutil.NoSuchProcess, psutil.AccessDenied):
        return False


def _handle_stop_signal(stop_event: threading.Event, signum: int, frame: Optional[FrameType]) -> None:
    try:
        signal_name = signal.Signals(signum).name
    except ValueError:
        signal_name = str(signum)

    logger.debug(f"Received signal {signal_name}")
    stop_event.set()


class SenderThread(Daemon):
    def __init__(
        self,
        api_token: str,
        family: str,
        operations_repository: OperationsRepository,
    ) -> None:
        super().__init__(name="SenderThread", sleep_time=SYNC_THREAD_SLEEP_TIME)

        self._api_token: str = api_token
        self._family: str = family
        self._operations_repository: OperationsRepository = operations_repository

        queued_range = operations_repository.get_operation_submission_sequence_id_range()
        if queued_range is None:
            last_queued_seq = SequenceId(-1)
        else:
            last_queued_seq = queued_range[1]
        self._last_queued_seq: SequenceId = last_queued_seq

        self._backend: Optional[ApiClient] = None
        self._metadata: Metadata = operations_repository.get_metadata()  # type: ignore

    @backoff.on_exception(backoff.expo, NeptuneRetryableError, max_time=HTTP_REQUEST_MAX_TIME_SECONDS)
    @with_api_errors_handling
    def submit(self, *, operation: RunOperation) -> Optional[SubmitResponse]:
        if self._backend is None:
            self._backend = ApiClient(api_token=self._api_token)

        response = self._backend.submit(operation=operation, family=self._family)

        status_code = response.status_code
        if status_code != 200:
            _raise_exception(status_code)

        return SubmitResponse.FromString(response.content)

    def work(self) -> None:
        try:
            max_operations_size = (
                MAX_REQUEST_SIZE_BYTES
                - len(self._metadata.run_id)
                - len(self._metadata.project)
                - 200  # 200 bytes for RunOperation overhead,
            )

            while operations := self._operations_repository.get_operations(
                from_exclusive=self._last_queued_seq, up_to_bytes=max_operations_size
            ):
                partitioned_operations = _partition_by_type_and_size(
                    operations, self._metadata.run_id, self._metadata.project, max_operations_size
                )

                logger.debug(
                    "Start: submit %d RunOperations. Last queued seq: #%d",
                    len(partitioned_operations),
                    self._last_queued_seq,
                )

                for run_operation, sequence_id, timestamp in partitioned_operations:
                    try:
                        request_ids: Optional[SubmitResponse] = self.submit(operation=run_operation)

                        if request_ids is None or not request_ids.request_ids:
                            raise NeptuneUnexpectedError("Server response is empty")

                        last_request_id = request_ids.request_ids[-1]

                        self._operations_repository.save_operation_submissions(
                            [
                                OperationSubmission(
                                    sequence_id=sequence_id,
                                    request_id=last_request_id,
                                    timestamp=int(timestamp.timestamp() * 1000),
                                )
                            ]
                        )

                        self._last_queued_seq = sequence_id
                    except NeptuneRetryableError as e:
                        self._operations_repository.save_errors([e], sequence_id=sequence_id)
                        # Sleep before retry
                        return

                logger.debug(
                    "Done: submit %d RunOperations. Last queued seq: #%d",
                    len(partitioned_operations),
                    self._last_queued_seq,
                )

        except NeptuneUnableToLogData:
            logger.warning("Failed to access the on disk operation log", exc_info=True)
        except Exception as e:
            self._operations_repository.save_errors([e])
            self.interrupt()
            raise NeptuneSynchronizationStopped() from e

    def close(self) -> None:
        if self._backend is not None:
            self._backend.close()


def _raise_exception(status_code: int) -> None:
    logger.warning("HTTP response error: %s", status_code)
    if status_code == 403:
        raise NeptuneUnauthorizedError()
    elif status_code == 408:
        raise NeptuneConnectionLostError()
    elif status_code == 429:
        raise NeptuneTooManyRequestsResponseError()
    elif status_code // 100 == 5:
        raise NeptuneInternalServerError()
    else:
        raise NeptuneUnexpectedResponseError()


def _partition_by_type_and_size(
    operations: list[Operation], run_id: str, project: str, max_batch_size: int
) -> list[tuple[RunOperation, SequenceId, datetime.datetime]]:
    grouped: list[list[Operation]] = []
    batch: list[Operation] = []
    batch_type: Optional[OperationType] = None
    batch_size = 0

    for op in operations:
        reset_batch = (
            # we don't mix operation types in a single batch
            op.operation_type != batch_type
            # only one CREATE_RUN per batch
            or batch_type == OperationType.CREATE_RUN
            # batch cannot be too big
            or batch_size + proto_encoded_bytes_field_size(op.operation_size_bytes) > max_batch_size
        )
        if reset_batch:
            if batch:
                grouped.append(batch)
            batch = []
            batch_type = op.operation_type
            batch_size = 0

        batch.append(op)
        batch_size += proto_encoded_bytes_field_size(op.operation_size_bytes)

    if batch:
        grouped.append(batch)

    def to_run_operation(ops: list[Operation]) -> tuple[RunOperation, SequenceId, datetime.datetime]:
        if ops[0].operation_type == OperationType.CREATE_RUN:
            return (
                RunOperation(project=project, run_id=run_id, create=ops[0].operation),
                ops[-1].sequence_id,
                ops[-1].ts,
            )
        else:
            snapshots = UpdateRunSnapshots(snapshots=[_op.operation for _op in ops])
            return RunOperation(project=project, run_id=run_id, update_batch=snapshots), ops[-1].sequence_id, ops[-1].ts

    return [(to_run_operation(ops)) for ops in grouped]


class StatusTrackingThread(Daemon):
    def __init__(
        self,
        api_token: str,
        project: str,
        operations_repository: OperationsRepository,
    ) -> None:
        super().__init__(name="StatusTrackingThread", sleep_time=STATUS_TRACKING_THREAD_SLEEP_TIME)

        self._api_token: str = api_token
        self._project: str = project
        self._operations_repository: OperationsRepository = operations_repository

        self._backend: Optional[ApiClient] = None

    def close(self) -> None:
        if self._backend is not None:
            self._backend.close()

    @backoff.on_exception(backoff.expo, NeptuneRetryableError, max_time=HTTP_REQUEST_MAX_TIME_SECONDS)
    @with_api_errors_handling
    def check_batch(self, *, request_ids: list[str]) -> Optional[BulkRequestStatus]:
        if self._backend is None:
            self._backend = ApiClient(api_token=self._api_token)

        response = self._backend.check_batch(request_ids=request_ids, project=self._project)

        status_code = response.status_code

        if status_code != 200:
            _raise_exception(status_code)

        return BulkRequestStatus.FromString(response.content)

    def work(self) -> None:
        try:
            while batch := self._operations_repository.get_operation_submissions(limit=MAX_REQUESTS_STATUS_BATCH_SIZE):
                request_ids = [element.request_id for element in batch]
                sequence_ids = [element.sequence_id for element in batch]

                try:
                    response = self.check_batch(request_ids=request_ids)
                    if response is None:
                        raise NeptuneUnexpectedError("Server response is empty")
                except NeptuneRetryableError as e:
                    self._operations_repository.save_errors([e])
                    # Small give up, sleep before retry
                    break

                processed_sequence_id, fatal_sync_error = None, None
                for request_status, request_sequence_id in zip(response.statuses, sequence_ids):
                    if any(code_status.code == Code.UNAVAILABLE for code_status in request_status.code_by_count):
                        logger.debug(f"Operation #{request_sequence_id} is not yet processed.")
                        # Request status not ready yet, sleep and retry
                        break

                    errors = [
                        code_to_exception(status.detail)
                        for status in request_status.code_by_count
                        if status.code != Code.OK and self._should_error_be_reported(status.detail)
                    ]

                    fatal_sync_error = next(filter(self._is_fatal_error, errors), None)
                    if fatal_sync_error is not None:
                        break

                    if errors:
                        self._operations_repository.save_errors(
                            errors=errors,
                            sequence_id=request_sequence_id,
                        )

                    processed_sequence_id = request_sequence_id

                if processed_sequence_id is not None:
                    logger.debug(f"Operations up to #{processed_sequence_id} are completed.")

                    # TODO: delete in a single transaction
                    self._operations_repository.delete_operations(up_to_seq_id=processed_sequence_id)
                    self._operations_repository.delete_operation_submissions(up_to_seq_id=processed_sequence_id)

                if fatal_sync_error is not None:
                    raise fatal_sync_error

                if processed_sequence_id is None:
                    # Sleep before retry
                    break
        except NeptuneUnableToLogData:
            logger.warning("Failed to access the on disk operation log", exc_info=True)
        except Exception as e:
            self._operations_repository.save_errors([e])
            self.interrupt()
            raise NeptuneSynchronizationStopped() from e

    @staticmethod
    def _is_fatal_error(ex: Exception) -> bool:
        return isinstance(ex, NeptuneProjectError) or isinstance(ex, NeptuneRunError)

    @staticmethod
    def _should_error_be_reported(ingest_code: IngestCode.ValueType) -> bool:
        return ingest_code not in {
            # We're hiding this error since historically neptune-scale prevented sending such values,
            # and now the backend supports them.
            # We only expect this error in the rare cases when the backend instance is outdated and the client is new
            #   - in such cases we don't want to spam the user with errors.
            # This code should be removed in the future when we can assume that all backend instances are updated.
            IngestCode.FLOAT_VALUE_NAN_INF_UNSUPPORTED,
        }


class FileUploaderThread(Daemon):
    def __init__(
        self,
        project: str,
        api_token: str,
        operations_repository: OperationsRepository,
        max_concurrent_uploads: Optional[int] = None,
    ) -> None:
        super().__init__(name="FileUploaderThread", sleep_time=1)

        self._project = project
        self._neptune_api_token = api_token
        self._operations_repository = operations_repository
        self._max_concurrent_uploads = max_concurrent_uploads or envs.get_positive_int(
            envs.MAX_CONCURRENT_FILE_UPLOADS, 50
        )
        self._upload_chunk_size = envs.get_positive_int("NEPTUNE_FILE_UPLOAD_CHUNK_SIZE", 16 * 1024 * 1024)
        logger.debug(f"{self._upload_chunk_size=}")
        # This is enforced by GCP, enforcing this for other providers is consistent.
        if self._upload_chunk_size % (256 * 1024) != 0:
            raise ValueError("NEPTUNE_FILE_UPLOAD_CHUNK_SIZE must be a multiple of 256 KiB")

        self._api_client: Optional[ApiClient] = None

        # AIO loop used for file uploads. It needs to be set as current event loop in the actual worker thread,
        # so we do it in work() when also initializing the API client.
        self._aio_loop: asyncio.AbstractEventLoop = asyncio.new_event_loop()

    def work(self) -> None:
        try:
            try:
                if self._api_client is None:
                    self._api_client = ApiClient(api_token=self._neptune_api_token)
                    asyncio.set_event_loop(self._aio_loop)

                while file_upload_requests := self._operations_repository.get_file_upload_requests(
                    self._max_concurrent_uploads
                ):
                    logger.debug(f"Have {len(file_upload_requests)} file upload requests to process")

                    file_sign_requests = [
                        FileSignRequest(
                            path=file.destination,
                            size=file.size_bytes,
                            permission="write",
                        )
                        for file in file_upload_requests
                    ]
                    storage_urls_all = fetch_file_storage_urls(self._api_client, self._project, file_sign_requests)

                    # Fan out file uploads as async tasks, and block until all tasks are done.
                    # Note that self._upload_file() should not raise an exception, as they are handled
                    # in the method itself. However, we still pass return_exceptions=True to make sure asyncio.gather()
                    # waits for all the tasks to finish regardless of any exceptions.
                    tasks = []
                    for file in file_upload_requests:
                        storage_urls = storage_urls_all[file.destination]
                        tasks.append(self._upload_file(file, storage_urls))

                    self._aio_loop.run_until_complete(asyncio.gather(*tasks, return_exceptions=True))

                    logger.debug("Upload tasks completed for the current batch")
            except NeptuneRetryableError as e:
                self._operations_repository.save_errors([e])
        except NeptuneUnableToLogData:
            logger.warning("Failed to access the on disk operation log", exc_info=True)
        except Exception as e:
            logger.error("Fatal error in file uploader thread", exc_info=e)
            self._operations_repository.save_errors([e])
            self.interrupt()
            raise NeptuneSynchronizationStopped() from e

    async def _upload_file(self, file: FileUploadRequest, storage_urls: StorageUrlResult) -> None:
        try:
            provider = storage_urls.provider
            if provider == Provider.AZURE:
                await upload_to_azure(
                    file.source_path, file.mime_type, storage_urls.url, chunk_size=self._upload_chunk_size
                )
            elif provider == Provider.GCP:
                await upload_to_gcp(
                    file.source_path, file.mime_type, storage_urls.url, chunk_size=self._upload_chunk_size
                )
            elif provider == Provider.AWS:
                if storage_urls.multipart_upload is None:
                    await upload_to_s3_single(file.source_path, file.mime_type, storage_urls.url)
                else:
                    multipart = storage_urls.multipart_upload
                    etags = await upload_to_s3_multipart(
                        file.source_path, file.mime_type, multipart.part_size, multipart.part_urls
                    )
                    complete_multipart_upload(
                        self._api_client,
                        project=self._project,
                        upload_id=multipart.upload_id,
                        destination=file.destination,
                        etags=etags,
                    )
            else:
                raise NeptuneUnexpectedError(f"Unsupported file storage provider: {provider}")

            if file.is_temporary:
                logger.debug(f"Removing temporary file {file.source_path}")
                Path(file.source_path).unlink(missing_ok=True)

            self._operations_repository.delete_file_upload_requests([file.sequence_id])
        except NeptuneRetryableError as e:
            self._operations_repository.save_errors([e])
        except Exception as e:
            # Fatal failure. Do not retry the file, but keep it on disk.
            logger.error(f"Error while uploading file {file.source_path}", exc_info=e)
            self._operations_repository.save_errors([NeptuneFileUploadError()])
            self._operations_repository.delete_file_upload_requests([file.sequence_id])

    def close(self) -> None:
        if self._api_client is not None:
            self._api_client.close()

        if self._aio_loop is not None:
            self._aio_loop.close()


@dataclass(frozen=True)
class MultipartUpload:
    upload_id: str
    part_size: int
    part_urls: list[str]


@dataclass(frozen=True)
class StorageUrlResult:
    provider: str
    url: str
    multipart_upload: Optional[MultipartUpload]


@backoff.on_exception(backoff.expo, NeptuneRetryableError, max_time=HTTP_REQUEST_MAX_TIME_SECONDS)
@with_api_errors_handling
def fetch_file_storage_urls(
    client: ApiClient, project: str, file_sign_requests: list[FileSignRequest]
) -> dict[str, StorageUrlResult]:
    """Fetch signed URLs for storing files. Returns a dict of target_path -> (provider, upload url)."""

    logger.debug("Fetching file storage urls")
    response = client.fetch_file_storage_urls(file_sign_requests=file_sign_requests, project=project)
    status_code = response.status_code
    if status_code // 100 != 2:
        logger.debug(f"{response.content=!r}")
        _raise_exception(status_code)

    parsed = response.parsed
    if parsed is None:
        raise NeptuneUnexpectedResponseError("Server response is empty")

    results = {}
    for file in parsed.files or []:
        multipart_upload = None
        if file.multipart:
            multipart_upload = MultipartUpload(
                upload_id=file.multipart.upload_id,
                part_size=file.multipart.part_size,
                part_urls=list(file.multipart.part_urls),
            )

        result = StorageUrlResult(provider=str(file.provider), url=file.url, multipart_upload=multipart_upload)
        results[file.path] = result
    return results


@backoff.on_exception(backoff.expo, NeptuneRetryableError, max_time=HTTP_REQUEST_MAX_TIME_SECONDS)
@with_api_errors_handling
def complete_multipart_upload(
    client: ApiClient, project: str, upload_id: str, destination: str, etags: Mapping[int, str]
) -> None:
    """Fetch signed URLs for storing files. Returns a dict of target_path -> (provider, upload url)."""

    logger.debug("Completing multipart upload")
    response = client.complete_multipart_upload(upload_id=upload_id, project=project, path=destination, etags=etags)
    status_code = response.status_code
    if status_code // 100 != 2:
        logger.debug(f"{response.content=!r}")
        _raise_exception(status_code)



from dataclasses import (
    dataclass,
    field,
)
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    TypeVar,
    Union,
)

if TYPE_CHECKING:
    try:
        import numpy as np  # noqa: F401
    except ImportError:
        np = None


@dataclass
class File:
    """
    Specifies details for a file being assigned or logged to an attribute.
    Used as a value in the dictionary passed to `Run.assign_files`
    when options beyond just the file source are needed.
    """

    # Source of the file content (path or binary file-like object).
    source: Union[str, Path, bytes]

    # Optional MIME type of the file (e.g., "image/png", "text/csv").
    mime_type: Optional[str] = field(default=None)

    # Optional size of the file in bytes.
    size: Optional[int] = field(default=None)

    # Optional destination path in object storage (relative to project namespace).
    destination: Optional[str] = field(default=None)


T = TypeVar("T")
ArrayLike = Union[list[T], "np.ndarray"]

try:
    import numpy as np

    _HAS_NUMPY = True
    # What types we accept in ArrayLike fields in Histogram
    _VALID_ARRAYLIKE_TYPES: tuple[Any, ...] = (list, np.ndarray)

    # If necessary, convert np.ndarray to plain python list for protobuf serialization
    def _as_list(arr: ArrayLike[T]) -> list[T]:
        return arr.tolist() if isinstance(arr, np.ndarray) and arr is not None else arr

except ImportError:
    _HAS_NUMPY = False
    _VALID_ARRAYLIKE_TYPES = (list,)

    def _as_list(arr: ArrayLike[T]) -> list[T]:
        return arr


@dataclass
class Histogram:
    """Represents a histogram with explicit bin edges.

    To specify the data distribution across bins, you can use either counts or densities. Note that n bins have n+1
    bin edges, so the length of the `bin_edges` argument should be one more than the length of the array
    that specifies the counts or densities.

    Use the histogram as a value in the dictionary passed to `Run.log_histograms()`.

    Args:
        bin_edges (Union[list[Union[float, int]], "np.ndarray"]): The bin edges of the histogram.
            Can't be empty. The maximum number of edges is 513.
        counts (Union[list[int], "np.ndarray"]): Number of data points that fall into each bin, as a 1D array.
        densities (Union[list[Union[float, int]], "np.ndarray"]): Probability density function values for each bin,
            as a 1D array.

    Examples:

        ```
        from neptune_scale.types import Histogram


        neptune_histogram = Histogram(bin_edges=[0, 1, 10, 1000], counts=[135, 289, 45])
        ```

        Using NumPy:

        ```
        import numpy as np


        a = np.arange(5)

        # Using sample count per bin
        counts, bin_edges = np.histogram(a, density=False)
        neptune_histogram = Histogram(bin_edges=bin_edges, counts=counts)

        # Using density values
        densities, bin_edges = np.histogram(a, density=True)
        neptune_histogram = Histogram(bin_edges=bin_edges, densities=densities)
        ```
    """

    bin_edges: ArrayLike[Union[float, int]]
    counts: Optional[ArrayLike[int]] = None
    densities: Optional[ArrayLike[Union[float, int]]] = None

    def bin_edges_as_list(self) -> list[Union[float, int]]:
        if not isinstance(self.bin_edges, _VALID_ARRAYLIKE_TYPES):
            raise TypeError(f"Bin edges must be of type list or np.ndarray, got {type(self.bin_edges)}")
        return _as_list(self.bin_edges)

    def counts_as_list(self) -> list[int]:
        if not isinstance(self.counts, _VALID_ARRAYLIKE_TYPES):
            raise TypeError(f"Counts must be of type list or np.ndarray, got {type(self.counts)}")
        return _as_list(self.counts)

    def densities_as_list(self) -> list[Union[float, int]]:
        if not isinstance(self.densities, _VALID_ARRAYLIKE_TYPES):
            raise TypeError(f"Densities must be of type list or np.ndarray, got {type(self.densities)}")
        return _as_list(self.densities)



import io
import sys
import threading
from datetime import (
    datetime,
    timedelta,
)
from typing import (
    Any,
    Callable,
    Optional,
    Union,
)

from neptune_scale.exceptions import NeptuneUnableToLogData
from neptune_scale.logging.logging_utils import (
    PartialLine,
    captured_data_to_lines,
    print_to_original_stderr,
    split_long_line,
)
from neptune_scale.sync.metadata_splitter import decompose_step
from neptune_scale.sync.parameters import MAX_STRING_SERIES_DATA_POINT_LENGTH
from neptune_scale.util import Daemon

STREAM_BUFFER_CHAR_CAPACITY = 100_000_000


class StreamWithMemory:
    def __init__(self, original_stream: io.TextIOBase):
        self._original_stream = original_stream
        self._fifo_buffer: list[tuple[datetime, str]] = []
        self._chars_in_buffer = 0
        self._lock = threading.RLock()
        self._subscribers_to_offset: dict[str, int] = {}

    def register_subscriber(self, subscriber_id: str) -> None:
        with self._lock:
            if subscriber_id not in self._subscribers_to_offset:
                self._subscribers_to_offset[subscriber_id] = len(self._fifo_buffer)

    def unregister_subscriber(self, subscriber_id: str) -> int:
        with self._lock:
            if subscriber_id in self._subscribers_to_offset:
                del self._subscribers_to_offset[subscriber_id]

            return len(self._subscribers_to_offset)

    def write(self, data: str) -> int:
        """
        Buffers the data with a timestamp for background processing AND
        writes the data immediately to the original stream.
        """
        ts = datetime.now()
        chars_written = self._original_stream.write(data)

        # Handle case where write() returns None (some stream implementations do this)
        if chars_written is None:
            chars_written = len(data)

        with self._lock:
            space_left_in_buffer = STREAM_BUFFER_CHAR_CAPACITY - self._chars_in_buffer
            if chars_written > space_left_in_buffer:
                # If the buffer is full, drop the oldest data
                # We always want to drop at least half of the data so that the amortized complexity remains O(1)
                chars_to_drop = max(chars_written, STREAM_BUFFER_CHAR_CAPACITY // 2)
                self._drop_first_chars_from_buffer(chars_to_drop)

            chars_to_save_in_buffer = min(chars_written, STREAM_BUFFER_CHAR_CAPACITY)
            self._fifo_buffer.append((ts, data[:chars_to_save_in_buffer]))
            self._chars_in_buffer += chars_to_save_in_buffer

        return chars_written

    def get_buffered_data(self, subscriber_id: str) -> list[tuple[datetime, str]]:
        """
        Atomically retrieves all buffered data for a specific subscriber (from its current offset till the end).
        """
        with self._lock:
            offset = self._subscribers_to_offset[subscriber_id]
            data_list = self._fifo_buffer[offset:]
            self._subscribers_to_offset[subscriber_id] = len(self._fifo_buffer)
            return data_list

    def _drop_first_chars_from_buffer(self, chars_to_drop: int) -> None:
        """
        Drops the first n entries from the buffer and adjust subscriber offsets accordingly.
        """
        with self._lock:
            chars_to_drop = min(chars_to_drop, self._chars_in_buffer)
            prefix_chars_count = 0
            entries_to_drop = 0
            while prefix_chars_count < chars_to_drop and entries_to_drop < len(self._fifo_buffer):
                prefix_chars_count += len(self._fifo_buffer[entries_to_drop][1])
                entries_to_drop += 1

            self._fifo_buffer = self._fifo_buffer[entries_to_drop:]
            self._chars_in_buffer -= prefix_chars_count

            for subscriber_id, offset in self._subscribers_to_offset.items():
                if offset >= entries_to_drop:
                    self._subscribers_to_offset[subscriber_id] = offset - entries_to_drop
                else:
                    self._subscribers_to_offset[subscriber_id] = 0

    def __getattr__(self, attr: str) -> Any:
        return getattr(self._original_stream, attr)


_lock = threading.RLock()
_stdout_with_memory: Optional[StreamWithMemory] = None
_stderr_with_memory: Optional[StreamWithMemory] = None
_original_stdout = sys.stdout
_original_stderr = sys.stderr


def _subscribe(subscriber_id: str) -> None:
    global _stdout_with_memory, _stderr_with_memory, _lock
    with _lock:
        if _stdout_with_memory is None:
            _stdout_with_memory = StreamWithMemory(sys.stdout)  # type: ignore
            sys.stdout = _stdout_with_memory
        if _stderr_with_memory is None:
            _stderr_with_memory = StreamWithMemory(sys.stderr)  # type: ignore
            sys.stderr = _stderr_with_memory

        _stdout_with_memory.register_subscriber(subscriber_id)
        _stderr_with_memory.register_subscriber(subscriber_id)


def _cancel_subscription(subscriber_id: str) -> None:
    global _stdout_with_memory, _stderr_with_memory, _lock
    with _lock:
        if _stdout_with_memory is not None:
            remaining_subscribers = _stdout_with_memory.unregister_subscriber(subscriber_id)
            if remaining_subscribers == 0:
                sys.stdout = _original_stdout
                _stdout_with_memory = None

        if _stderr_with_memory is not None:
            remaining_subscribers = _stderr_with_memory.unregister_subscriber(subscriber_id)
            if remaining_subscribers == 0:
                sys.stderr = _original_stderr
                _stderr_with_memory = None


class StepTracker:
    def __init__(self, initial_step: Union[int, float]) -> None:
        """Track a float step value with a precision of 1e-6."""
        self.whole, self.micro = decompose_step(initial_step)
        # Always start with the subsequent step
        self.increment()

    @property
    def value(self) -> float:
        return self._value

    def increment(self) -> float:
        """Increment the step and return the new value."""
        self.micro += 1
        if self.micro >= 1_000_000:
            self.micro = 0
            self.whole += 1

        self._value = self.whole + self.micro / 1_000_000
        return self._value


class ConsoleLogCaptureThread(Daemon):
    def __init__(
        self,
        *,
        run_id: str,
        runtime_namespace: str,
        initial_step: Union[int, float],
        logs_flush_frequency_sec: float,
        logs_sink: Callable[[dict[str, str], Union[float, int], Optional[datetime]], None],
    ) -> None:
        super().__init__(logs_flush_frequency_sec, "ConsoleLogCapture")

        self._logs_sink = logs_sink
        self._run_id = run_id

        _subscribe(run_id)

        self._stdout_partial_line = PartialLine()
        self._stdout_attribute = f"{runtime_namespace}/stdout"
        self._stdout_step = StepTracker(initial_step)

        self._stderr_partial_line = PartialLine()
        self._stderr_attribute = f"{runtime_namespace}/stderr"
        self._stderr_step = StepTracker(initial_step)

    def work(self) -> None:
        self._process_captured_data(max_delay_before_flush=timedelta(seconds=5))

    def close(self) -> None:
        self._process_captured_data(max_delay_before_flush=timedelta(seconds=0))
        _cancel_subscription(self.name)

    def _process_captured_data(self, max_delay_before_flush: timedelta) -> None:
        try:
            assert _stdout_with_memory is not None
            assert _stderr_with_memory is not None

            self._process_captured_data_single_stream(
                self._stdout_attribute,
                self._stdout_partial_line,
                self._stdout_step,
                _stdout_with_memory.get_buffered_data(self._run_id),
                max_delay_before_flush,
            )

            self._process_captured_data_single_stream(
                self._stderr_attribute,
                self._stderr_partial_line,
                self._stderr_step,
                _stderr_with_memory.get_buffered_data(self._run_id),
                max_delay_before_flush,
            )

        except NeptuneUnableToLogData as e:
            print_to_original_stderr(f"Unable to log data from stdout/stderr: {e}")

        except Exception as e:
            _cancel_subscription(self.name)
            raise e

    def _process_captured_data_single_stream(
        self,
        attribute_name: str,
        partial_line: PartialLine,
        step: StepTracker,
        data: list[tuple[datetime, str]],
        max_delay_before_flush: timedelta,
    ) -> None:
        for ts, line in captured_data_to_lines(partial_line, data, max_delay_before_flush):
            for short_line in split_long_line(line, MAX_STRING_SERIES_DATA_POINT_LENGTH):
                self._logs_sink({attribute_name: short_line}, step.value, ts)
                step.increment()



from __future__ import annotations

__all__ = ("LagTracker",)

import time
from collections.abc import Callable

from neptune_scale.sync.operations_repository import OperationsRepository
from neptune_scale.sync.parameters import LAG_TRACKER_THREAD_SLEEP_TIME
from neptune_scale.util import Daemon


class LagTracker(Daemon):
    def __init__(
        self,
        operations_repository: OperationsRepository,
        async_lag_threshold: float,
        on_async_lag_callback: Callable[[], None],
    ) -> None:
        super().__init__(name="LagTracker", sleep_time=LAG_TRACKER_THREAD_SLEEP_TIME)

        self._operations_repository = operations_repository
        self._async_lag_threshold: float = async_lag_threshold
        self._on_async_lag_callback: Callable[[], None] = on_async_lag_callback

        self._callback_triggered: bool = False

    def work(self) -> None:
        oldest_queued_timestamp = self._operations_repository.get_operations_min_timestamp()
        current_timestamp = time.time()

        if (
            oldest_queued_timestamp is not None
            and current_timestamp - oldest_queued_timestamp.timestamp() > self._async_lag_threshold
        ):
            if not self._callback_triggered:
                self._on_async_lag_callback()
                self._callback_triggered = True
        else:
            self._callback_triggered = False



import hashlib
import mimetypes
import pathlib
from typing import (
    Optional,
    Union,
    cast,
)

import filetype

from neptune_scale.sync.parameters import MAX_FILE_DESTINATION_LENGTH
from neptune_scale.util import get_logger

__all__ = (
    "guess_mime_type_from_file",
    "guess_mime_type_from_bytes",
    "generate_destination",
)

DEFAULT_MIME_TYPE = "application/octet-stream"
logger = get_logger()

EXTRA_TYPES = {
    ".yaml": "application/yaml",
    ".yml": "application/yaml",
    ".r": "text/x-r",
    ".ipynb": "application/x-ipynb+json",
    ".md": "text/markdown",
    ".sql": "application/x-sql",
    ".webp": "image/webp",
}


def guess_mime_type_from_file(local_path: Union[pathlib.Path, str], destination: Optional[str] = None) -> Optional[str]:
    """Guess mime type by local file path and the destination path. In case of an error, return None.

    We return None instead of DEFAULT_MIME_TYPE under the assumption that an error here also means
    that the file is inaccessible or not found, thus the upload will fail.
    """
    try:
        for extra_ext, extra_mime in EXTRA_TYPES.items():
            if mimetypes.guess_type(f"dummy{extra_ext}")[0] is None:
                mimetypes.add_type(extra_mime, extra_ext)

        if mime := mimetypes.guess_type(local_path)[0]:
            return mime

        if destination:
            if mime := mimetypes.guess_type(destination)[0]:
                return mime

        if mime := filetype.guess_mime(local_path):
            return cast(str, mime)

        return DEFAULT_MIME_TYPE
    except Exception as e:
        logger.warning(f"Error determining mime type for {local_path}: {e}")
        return None


def guess_mime_type_from_bytes(data: bytes, destination: Optional[str] = None) -> str:
    """Guess mime type by providing a buffer and the target path"""
    try:
        if destination:
            if mime := mimetypes.guess_type(destination)[0]:
                return mime

        if mime := filetype.guess_mime(data):
            return cast(str, mime)

        return DEFAULT_MIME_TYPE
    except Exception as e:
        logger.warning(f"Error determining mime type for the provided buffer, defaulting to {DEFAULT_MIME_TYPE}: {e}")
        return DEFAULT_MIME_TYPE


# Maximum lengths of various components of the file destination.
MAX_RUN_ID_COMPONENT_LENGTH = 300
MAX_ATTRIBUTE_PATH_COMPONENT_LENGTH = 280
MAX_SERIES_STEP_PATH_COMPONENT_LENGTH = 19
NO_STEP_PLACEHOLDER = "none"
MAX_FILENAME_PATH_COMPONENT_LENGTH = 180
MAX_FILENAME_EXTENSION_LENGTH = 18

# Format: "run_id/series_ttribute_path/series_step/file.txt", with +3 to account for the "/" separators.
assert (
    MAX_RUN_ID_COMPONENT_LENGTH
    + MAX_ATTRIBUTE_PATH_COMPONENT_LENGTH
    + MAX_SERIES_STEP_PATH_COMPONENT_LENGTH
    + MAX_FILENAME_PATH_COMPONENT_LENGTH
    + MAX_FILENAME_EXTENSION_LENGTH
    + 3
) == MAX_FILE_DESTINATION_LENGTH


def _digest_suffix(string: str) -> str:
    digest = hashlib.blake2b(string.encode("utf-8"), digest_size=8).hexdigest()
    return "-" + digest


_DISALLOWED_CHARS_REPLACEMENT = str.maketrans({char: "_" for char in "/."})


def _sanitize_and_trim(input_str: str, max_length: int, force_suffix: bool) -> str:
    sanitized = input_str.translate(_DISALLOWED_CHARS_REPLACEMENT)

    suffix = _digest_suffix(input_str) if force_suffix or len(sanitized) > max_length else ""
    return sanitized[: max_length - len(suffix)] + suffix


def generate_destination(run_id: str, attribute_name: str, filename: str, step: Optional[Union[float, int]]) -> str:
    """
    Generate a path under which a file should be saved in the storage.

    The path is generated in the format:
        <run_id>/<attribute_name>/<filename>

    The path is guaranteed not to exceed the max length. If necessary, path
    components are truncated to fit their maximum allowed lengths.

    All path components have "." and "/" replaced with "_". Run id an attribute path
    always have a hash digest appended.

    Filenames have "." and "/" characters replaced with "_", except for "." in the extension.
    The extension is truncated to a maximum length of 18 characters, and
    the file stem, if necessary, is truncated with digest appended.
    """

    run_id = _sanitize_and_trim(run_id, MAX_RUN_ID_COMPONENT_LENGTH, force_suffix=True)
    attribute_name = _sanitize_and_trim(attribute_name, MAX_ATTRIBUTE_PATH_COMPONENT_LENGTH, force_suffix=True)

    # steps longer that 12 + 6 digits should be rejected by the server, but let's be safe and truncate them here
    if step is None:
        series_step = NO_STEP_PLACEHOLDER
    else:
        series_step = _sanitize_and_trim(f"{step:019.6f}", MAX_SERIES_STEP_PATH_COMPONENT_LENGTH, force_suffix=False)

    # Sanitize the filename. Truncate if necessary, keeping extension (truncated) if present.
    path = pathlib.Path(filename)
    extension = path.suffix  # includes the dot (.txt)
    filename_no_ext = path.name[: -len(extension)] if extension else path.name
    filename_no_ext = _sanitize_and_trim(filename_no_ext, MAX_FILENAME_PATH_COMPONENT_LENGTH, force_suffix=False)
    filename = filename_no_ext + extension[:MAX_FILENAME_EXTENSION_LENGTH]

    return f"{run_id}/{attribute_name}/{series_step}/{filename}"



__all__ = ["RepositoryInfo", "read_repository_info"]

import pathlib
from dataclasses import dataclass
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Optional,
)

from neptune_scale.util.logger import get_logger
from neptune_scale.vendor.lib_programname import (
    empty_path,
    get_path_executed_script,
)

if TYPE_CHECKING:
    import git

logger = get_logger()


@dataclass
class RepositoryInfo:
    commit_id: str
    commit_message: str
    commit_author_name: str
    commit_author_email: str
    commit_date: datetime
    dirty: bool
    branch: Optional[str]
    remotes: dict[str, str]
    run_command: Optional[str]
    entry_point_path: Optional[pathlib.Path]
    head_diff_content: Optional[bytes]
    upstream_diff_commit_id: Optional[str]
    upstream_diff_content: Optional[bytes]


def read_repository_info(
    path: Optional[pathlib.Path], run_command: bool, entry_point: bool, head_diff: bool, upstream_diff: bool
) -> Optional[RepositoryInfo]:
    git_repo = _get_git_repo(path)
    if git_repo is None:
        return None

    try:
        head_commit = git_repo.head.commit
    except ValueError:
        # Happens when the repository is empty (no commits)
        return None

    is_dirty = git_repo.is_dirty(index=False)
    active_branch = _get_active_branch(git_repo)
    remotes = {remote.name: remote.url for remote in git_repo.remotes}

    run_command_content = None
    if run_command:
        run_command_content = _get_run_command()

    entry_point_path = None
    if entry_point:
        entry_point_path = _get_entrypoint_path()

    head_diff_content = None
    if head_diff:
        head_diff_content = _read_diff(git_repo, git_repo.head.name)

    upstream_sha = None
    upstream_diff_content = None
    if upstream_diff:
        upstream_commit = _get_relevant_upstream_commit(git_repo)
        if upstream_commit:
            upstream_sha = upstream_commit.hexsha
            upstream_diff_content = _read_diff(git_repo, upstream_sha)

    return RepositoryInfo(
        commit_id=head_commit.hexsha,
        commit_message=head_commit.message,
        commit_author_name=head_commit.author.name,
        commit_author_email=head_commit.author.email,
        commit_date=head_commit.committed_datetime,
        dirty=is_dirty,
        branch=active_branch,
        remotes=remotes,
        entry_point_path=entry_point_path,
        head_diff_content=head_diff_content,
        upstream_diff_commit_id=upstream_sha,
        upstream_diff_content=upstream_diff_content,
        run_command=run_command_content,
    )


def _get_git_repo(path: Optional[pathlib.Path]) -> Optional["git.Repo"]:
    if path is not None:
        path = path.resolve()

    try:
        import git

        try:
            return git.Repo(path, search_parent_directories=True)
        except (git.exc.NoSuchPathError, git.exc.InvalidGitRepositoryError):
            return None
    except ImportError:
        logger.warning("GitPython could not be initialized")
        return None


def _get_active_branch(repo: "git.Repo") -> str:
    active_branch = ""
    try:
        active_branch = repo.active_branch.name
    except TypeError as e:
        if str(e.args[0]).startswith("HEAD is a detached symbolic reference as it points to"):
            active_branch = "Detached HEAD"
    return active_branch


def _read_diff(repo: "git.Repo", commit_ref: str) -> Optional[bytes]:
    try:
        from git.exc import GitCommandError

        try:
            diff = repo.git.diff(commit_ref, index=False, no_ext_diff=True)

            if not isinstance(diff, str):
                return None

            # add a newline at the end (required to be a valid `patch` file)
            if diff and diff[-1] != "\n":
                diff += "\n"

            return diff.encode("utf-8")
        except GitCommandError:
            return None
    except ImportError:
        return None


def _get_relevant_upstream_commit(repo: "git.Repo") -> Optional["git.Commit"]:
    try:
        tracking_branch = repo.active_branch.tracking_branch()
    except (TypeError, ValueError):
        return None

    if tracking_branch:
        return tracking_branch.commit

    return _search_for_most_recent_ancestor(repo)


def _search_for_most_recent_ancestor(repo: "git.Repo") -> Optional["git.Commit"]:
    most_recent_ancestor: Optional[git.Commit] = None

    try:
        from git.exc import GitCommandError

        try:
            for branch in repo.heads:
                tracking_branch = branch.tracking_branch()
                if tracking_branch:
                    for ancestor in repo.merge_base(repo.head, tracking_branch.commit):
                        if not most_recent_ancestor or repo.is_ancestor(most_recent_ancestor, ancestor):
                            most_recent_ancestor = ancestor
        except GitCommandError:
            pass
    except ImportError:
        return None

    return most_recent_ancestor


def _get_entrypoint_path() -> Optional[pathlib.Path]:
    if _is_ipython():
        return None

    entrypoint_path = get_path_executed_script()
    if not isinstance(entrypoint_path, pathlib.Path):
        return None
    if entrypoint_path == empty_path:
        return None
    if not entrypoint_path.is_file():
        return None

    return entrypoint_path


def _is_ipython() -> bool:
    try:
        import IPython

        ipython = IPython.core.getipython.get_ipython()
        return ipython is not None
    except ImportError:
        return False


def _get_run_command() -> str:
    import psutil

    args = psutil.Process().cmdline()
    return " ".join(args)



import asyncio
from pathlib import Path

import aiofiles
import backoff
import httpx
from httpx import AsyncClient

from neptune_scale.exceptions import NeptuneFileUploadTemporaryError
from neptune_scale.sync.parameters import (
    HTTP_CLIENT_NETWORKING_TIMEOUT,
    HTTP_REQUEST_MAX_TIME_SECONDS,
)
from neptune_scale.util import get_logger

__all__ = ["upload_to_s3_single", "upload_to_s3_multipart"]

logger = get_logger()


async def upload_to_s3_single(file_path: str, content_type: str, signed_url: str) -> None:
    """
    Upload a file to S3 using a signed URL. The upload is done in a single part and is retried in case of a failure.

    Raises NeptuneFileUploadTemporaryError if a retryable error happens, otherwise any other non-retryable exception
    that occurs.
    """

    logger.debug(f"Starting upload to S3: {file_path}, {content_type=}")

    try:
        async with AsyncClient(timeout=httpx.Timeout(timeout=HTTP_CLIENT_NETWORKING_TIMEOUT)) as client:
            async with aiofiles.open(file_path, "rb") as file:
                file_size = Path(file_path).stat().st_size
                if file_size == 0:
                    await _upload_content(client, signed_url, b"", content_type)
                    return

                content = await file.read()
                if not content:
                    raise OSError("File truncated during upload")

                await _upload_content(client, signed_url, content, content_type)
    except httpx.RequestError as e:
        logger.debug(f"Temporary error while uploading {file_path}: {e}")
        raise NeptuneFileUploadTemporaryError() from e
    except httpx.HTTPStatusError as e:
        logger.debug(f"HTTP {e.response.status_code} error while uploading {file_path}: {e}, {e.response.content=!r}")
        if _is_retryable_httpx_error(e):
            raise NeptuneFileUploadTemporaryError() from e
        else:
            raise

    logger.debug(f"Finished upload to S3: {file_path}")


async def upload_to_s3_multipart(
    file_path: str, content_type: str, part_size: int, part_urls: list[str]
) -> dict[int, str]:
    """
    Upload a file to S3 using a signed URL. The upload uses multiple signed urls for each part and finishes the upload
    by calling the complete multipart upload endpoint.

    Raises NeptuneFileUploadTemporaryError if a retryable error happens, otherwise any other non-retryable exception
    that occurs.
    """

    logger.debug(f"Starting upload to S3: {file_path}, {content_type=}")

    async def upload_part(part_ix: int, part_url: str, content: bytes) -> tuple[int, str]:
        etag = await _upload_content(client, part_url, content, content_type)
        return part_ix, etag

    try:
        async with AsyncClient(timeout=httpx.Timeout(timeout=HTTP_CLIENT_NETWORKING_TIMEOUT)) as client:
            file_size = Path(file_path).stat().st_size
            if file_size == 0:
                raise ValueError("Cannot upload an empty file with multipart upload")

            async with aiofiles.open(file_path, "rb") as file:
                tasks = []
                for part_ix, part_url in enumerate(part_urls, start=1):
                    content = await file.read(part_size)

                    if not content:
                        raise OSError("File truncated during upload")

                    tasks.append(upload_part(part_ix, part_url, content))

                results = await asyncio.gather(*tasks)
                etags = dict(results)

        logger.debug(f"Finished upload to S3: {file_path}")
        return etags

    except httpx.RequestError as e:
        logger.debug(f"Temporary error while uploading {file_path}: {e}")
        raise NeptuneFileUploadTemporaryError() from e
    except httpx.HTTPStatusError as e:
        logger.debug(f"HTTP {e.response.status_code} error while uploading {file_path}: {e}, {e.response.content=!r}")
        if _is_retryable_httpx_error(e):
            raise NeptuneFileUploadTemporaryError() from e
        else:
            raise


def _is_retryable_httpx_error(exc: Exception) -> bool:
    """Used to determine if an error is retryable. The standard retry policy is as defined in:
    https://github.com/boto/botocore/blob/master/botocore/retries/standard.py
    or https://boto3.amazonaws.com/v1/documentation/api/latest/guide/retries.html
    It might be worth considering to use the same policy for S3 uploads, but for now we retry on:
    - All network-related errors
    - HTTP 5xx errors
    - HTTP 429 Too Many Requests
    """
    if isinstance(exc, httpx.RequestError):
        return True
    if isinstance(exc, httpx.HTTPStatusError):
        status_code = exc.response.status_code
        return status_code in (429,) or status_code // 100 == 5

    return False


@backoff.on_predicate(backoff.expo, _is_retryable_httpx_error, max_time=HTTP_REQUEST_MAX_TIME_SECONDS)
async def _upload_content(client: AsyncClient, session_uri: str, content: bytes, content_type: str) -> str:
    # The docs at https://docs.aws.amazon.com/AmazonS3/latest/userguide/PresignedUrlUploadObject.html
    # provide Content-Type as the only header
    headers = {"Content-Type": content_type}

    response = await client.put(session_uri, headers=headers, content=content)
    response.raise_for_status()

    etag = response.headers.get("ETag", "")
    if not etag:
        raise ValueError("ETag header missing in S3 upload response")
    return str(etag).strip('"')



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from datetime import datetime
from typing import (
    Any,
    Optional,
    TypeVar,
    Union,
)

from google.protobuf.timestamp_pb2 import Timestamp
from more_itertools import peekable
from neptune_api.proto.neptune_pb.ingest.v1.common_pb2 import (
    SET_OPERATION,
    FileRef,
)
from neptune_api.proto.neptune_pb.ingest.v1.common_pb2 import Histogram as ProtobufHistogram
from neptune_api.proto.neptune_pb.ingest.v1.common_pb2 import (
    Preview,
    Step,
    StringSet,
    UpdateRunSnapshot,
    Value,
)

from neptune_scale.exceptions import NeptuneUnableToLogData
from neptune_scale.sync.parameters import (
    MAX_ATTRIBUTE_PATH_LENGTH,
    MAX_FILE_DESTINATION_LENGTH,
    MAX_FILE_MIME_TYPE_LENGTH,
    MAX_HISTOGRAM_BIN_EDGES,
    MAX_SINGLE_OPERATION_SIZE_BYTES,
    MAX_STRING_SERIES_DATA_POINT_LENGTH,
)
from neptune_scale.sync.size_util import (
    SINGLE_FLOAT_VALUE_SIZE,
    proto_string_size,
)
from neptune_scale.types import Histogram
from neptune_scale.util import (
    envs,
    get_logger,
)

__all__ = (
    "FileRefData",
    "MetadataSplitter",
    "Metrics",
    "datetime_to_proto",
    "make_step",
    "histograms_to_update_run_snapshots",
    "string_series_to_update_run_snapshots",
    "sanitize_attribute_path",
)


logger = get_logger()

T = TypeVar("T")


INVALID_VALUE_ACTION = envs.get_option(envs.LOG_FAILURE_ACTION, ("drop", "raise"), "drop")


@dataclass(frozen=True)
class FileRefData:
    """Passed between Run and MetadataSplitter for file uploads"""

    # py39 does not support slots=True on @dataclass
    __slots__ = ("destination", "mime_type", "size_bytes")

    destination: str
    mime_type: str
    size_bytes: int


@dataclass
class Metrics:
    """Class representing a set of metrics at a single step"""

    data: dict[str, Union[float, int]]
    preview: bool = False
    preview_completion: Optional[float] = None


class MetadataSplitter(Iterator[UpdateRunSnapshot]):
    def __init__(
        self,
        *,
        project: str,
        run_id: str,
        timestamp: datetime,
        step: Optional[Union[float, int]],
        configs: Optional[dict[str, Union[float, bool, int, str, datetime, list, set, tuple]]],
        metrics: Optional[Metrics],
        files: Optional[dict[str, FileRefData]],
        file_series: Optional[dict[str, FileRefData]],
        add_tags: Optional[dict[str, Union[list[str], set[str], tuple[str]]]],
        remove_tags: Optional[dict[str, Union[list[str], set[str], tuple[str]]]],
        max_message_bytes_size: int = MAX_SINGLE_OPERATION_SIZE_BYTES,
    ):
        self._timestamp = datetime_to_proto(timestamp)
        self._project = project
        self._run_id = run_id

        self._step = make_step(step) if step is not None else None
        self._metrics = peekable(self._stream_metrics(metrics.data)) if metrics is not None else None
        self._preview = _make_preview_from_metrics(metrics) if metrics else None

        self._configs = peekable(self._stream_configs(configs)) if configs else None
        self._add_tags = peekable(self._stream_tags(add_tags)) if add_tags else None
        self._remove_tags = peekable(self._stream_tags(remove_tags)) if remove_tags else None
        self._files = peekable(self._stream_files(files)) if files else None
        self._file_series = peekable(self._stream_files(file_series)) if file_series else None

        self._max_update_bytes_size = max_message_bytes_size

    def __iter__(self) -> MetadataSplitter:
        return self

    def __next__(self) -> UpdateRunSnapshot:
        if (
            not bool(self._configs)
            and not bool(self._metrics)
            and not bool(self._add_tags)
            and not bool(self._remove_tags)
            and not bool(self._files)
            and not bool(self._file_series)
        ):
            raise StopIteration

        update = UpdateRunSnapshot(step=self._step, timestamp=self._timestamp, preview=self._preview)
        size = update.ByteSize()

        size = self.populate_assign(
            update=update,
            assets=self._configs,
            size=size,
        )
        size = self.populate_append_metrics(
            update=update,
            assets=self._metrics,
            size=size,
        )
        size = self.populate_tags(
            update=update,
            assets=self._add_tags,
            operation=SET_OPERATION.ADD,
            size=size,
        )
        size = self.populate_tags(
            update=update,
            assets=self._remove_tags,
            operation=SET_OPERATION.REMOVE,
            size=size,
        )
        size = self.populate_assign(
            update=update,
            assets=self._files,
            size=size,
        )
        size = self.populate_append(
            update=update,
            assets=self._file_series,
            size=size,
        )

        return update

    def populate_assign(
        self,
        update: UpdateRunSnapshot,
        assets: Optional[peekable[tuple[str, Value]]],
        size: int,
    ) -> int:
        if assets is None:
            return size

        while size < self._max_update_bytes_size:
            try:
                key, value = assets.peek()
            except StopIteration:
                break

            new_size = size + proto_string_size(key) + value.ByteSize() + 6
            if new_size > self._max_update_bytes_size:
                break

            update.assign[key].MergeFrom(value)
            size, _ = new_size, next(assets)

        return size

    def populate_append(
        self,
        update: UpdateRunSnapshot,
        assets: Optional[peekable[Value]],
        size: int,
    ) -> int:
        if assets is None:
            return size

        while size < self._max_update_bytes_size:
            try:
                key, value = assets.peek()
            except StopIteration:
                break

            new_size = size + proto_string_size(key) + value.ByteSize() + 6
            if new_size > self._max_update_bytes_size:
                break

            update.append[key].MergeFrom(value)
            size, _ = new_size, next(assets)

        return size

    def populate_append_metrics(
        self,
        update: UpdateRunSnapshot,
        assets: Optional[peekable[tuple[str, float]]],
        size: int,
    ) -> int:
        if assets is None:
            return size

        while size < self._max_update_bytes_size:
            try:
                key, value = assets.peek()
            except StopIteration:
                break

            new_size = size + proto_string_size(key) + SINGLE_FLOAT_VALUE_SIZE + 6
            if new_size > self._max_update_bytes_size:
                break

            update.append[key].float64 = value
            size, _ = new_size, next(assets)

        return size

    def populate_tags(
        self, update: UpdateRunSnapshot, assets: Optional[peekable[Any]], operation: SET_OPERATION.ValueType, size: int
    ) -> int:
        if assets is None:
            return size

        while size < self._max_update_bytes_size:
            try:
                key, values = assets.peek()
            except StopIteration:
                break

            if not isinstance(values, peekable):
                values = peekable(values)

            is_full = False
            new_size = size + proto_string_size(key) + 6
            for value in values:
                tag_size = proto_string_size(value) + 6
                if new_size + tag_size > self._max_update_bytes_size:
                    values.prepend(value)
                    is_full = True
                    break

                update.modify_sets[key].string.values[value] = operation
                new_size += tag_size

            size, _ = new_size, next(assets)

            if is_full:
                assets.prepend((key, list(values)))
                break

        return size

    def _stream_metrics(self, metrics: dict[str, float]) -> Iterator[tuple[str, float]]:
        for key, value in _validate_paths(metrics):
            try:
                value = float(value)
            except (ValueError, TypeError, OverflowError):
                _warn_or_raise_on_invalid_value(f"Metrics' values must be float or int (got `{key}`:`{value}`)")
                continue

            yield key, value

    def _stream_configs(
        self, configs: dict[str, Union[float, bool, int, str, datetime, list, set, tuple]]
    ) -> Iterator[tuple[str, Value]]:
        _is_instance = isinstance  # local binding, faster in tight loops
        for key, value in _validate_paths(configs):
            if _is_instance(value, float):
                yield key, Value(float64=value)
            elif _is_instance(value, bool):
                yield key, Value(bool=value)
            elif _is_instance(value, int):
                yield key, Value(int64=value)
            elif _is_instance(value, str):
                yield key, Value(string=value)
            elif _is_instance(value, datetime):
                yield key, Value(timestamp=datetime_to_proto(value))  # type: ignore
            elif _is_instance(value, (list, set, tuple)):
                yield key, Value(string_set=StringSet(values=value))
            else:
                _warn_or_raise_on_invalid_value(
                    f"Config values must be float, bool, int, str, datetime, list, set or tuple "
                    f"(got `{key}`:`{value}`)"
                )
                continue

    def _stream_tags(
        self, tags: dict[str, Union[list[str], set[str], tuple[str]]]
    ) -> Iterator[tuple[str, Union[list[str], set[str], tuple[str]]]]:
        accepted_tag_collection_types = (list, set, tuple)
        _is_instance = isinstance  # local binding, faster in tight loops
        for key, values in _validate_paths(tags):
            if not _is_instance(values, accepted_tag_collection_types) or any(
                not _is_instance(tag, str) for tag in values
            ):
                _warn_or_raise_on_invalid_value(
                    f"Tags must be a list, set or tuple of strings (got `{key}`:`{values}`)"
                )
                continue

            yield key, values

    def _stream_files(self, files: dict[str, FileRefData]) -> Iterator[tuple[str, Value]]:
        _is_instance = isinstance
        for attr_name, file in _validate_paths(files):
            if len(file.destination) > MAX_FILE_DESTINATION_LENGTH:
                _warn_or_raise_on_invalid_value(
                    f"File destination must be a string of at most {MAX_FILE_DESTINATION_LENGTH} characters"
                    f"(got `{file.destination}` for {attr_name}`)"
                )
                continue

            if len(file.mime_type) > MAX_FILE_MIME_TYPE_LENGTH:
                _warn_or_raise_on_invalid_value(
                    f"File mime type must be a string of at most {MAX_FILE_MIME_TYPE_LENGTH} characters"
                    f" (got `{file.mime_type}` for `{attr_name}`)"
                )
                continue

            yield (
                attr_name,
                Value(file_ref=FileRef(path=file.destination, mime_type=file.mime_type, size_bytes=file.size_bytes)),
            )


def sanitize_attribute_path(param: str) -> str:
    if len(param) > MAX_ATTRIBUTE_PATH_LENGTH:
        param = param[:MAX_ATTRIBUTE_PATH_LENGTH]
    return param


def _validate_paths(fields: dict[str, T]) -> Iterator[tuple[str, T]]:
    # local bindings, faster in tight loops
    _is_instance = isinstance
    __is_over_utf8_bytes_limit = _is_over_utf8_bytes_limit
    _max_length = MAX_ATTRIBUTE_PATH_LENGTH

    for key, value in fields.items():
        if not _is_instance(key, str):
            _warn_or_raise_on_invalid_value(f"Field paths must be strings (got `{key}`)")
            continue

        if __is_over_utf8_bytes_limit(key, _max_length):
            _warn_or_raise_on_invalid_value(
                f"Field paths must be less than {_max_length} bytes when UTF-8 encoded (got `{key}`)"
            )
            continue

        yield key, value


def _make_preview_from_metrics(metrics: Metrics) -> Optional[Preview]:
    if not metrics.preview:
        return None
    # let backend default completion
    if metrics.preview_completion is not None:
        return Preview(is_preview=True, completion_ratio=metrics.preview_completion)
    return Preview(is_preview=True)


def string_series_to_update_run_snapshots(
    string_series: Optional[dict[str, str]],
    step: Optional[Union[float, int]],
    timestamp: datetime,
    max_size: int = MAX_SINGLE_OPERATION_SIZE_BYTES,
) -> Iterator[UpdateRunSnapshot]:
    if not string_series:
        return
    assert step is not None, "Step must be provided when string series are present"

    stream = peekable(_stream_string_series(string_series))
    step = make_step(step)
    timestamp = datetime_to_proto(timestamp)

    # Local bindings for faster name lookups
    _proto_string_size = proto_string_size
    _peek_stream = stream.peek
    while stream:
        update = UpdateRunSnapshot(step=step, timestamp=timestamp)

        size = 0
        while size < max_size:
            try:
                key, value = _peek_stream()
            except StopIteration:
                break

            new_size = size + _proto_string_size(key) + _proto_string_size(value) + 6
            if new_size > max_size:
                break

            update.append[key].string = value
            size, _ = new_size, next(stream)

        yield update


def _stream_string_series(string_series: dict[str, str]) -> Iterator[tuple[str, str]]:
    # local bindings, faster in tight loops
    _is_instance = isinstance
    __is_over_utf8_bytes_limit = _is_over_utf8_bytes_limit
    _max_length = MAX_STRING_SERIES_DATA_POINT_LENGTH

    for key, value in _validate_paths(string_series):
        if not _is_instance(value, str):
            _warn_or_raise_on_invalid_value(f"String series values must be strings (got `{key}`:`{value}`)")
            continue

        if __is_over_utf8_bytes_limit(value, _max_length):
            _warn_or_raise_on_invalid_value(
                f"String series values must be less than {_max_length} bytes when UTF-8 encoded"
            )
            continue

        yield key, value


def histograms_to_update_run_snapshots(
    histograms: Optional[dict[str, Histogram]],
    step: Optional[Union[float, int]],
    timestamp: datetime,
    max_size: int = MAX_SINGLE_OPERATION_SIZE_BYTES,
) -> Iterator[UpdateRunSnapshot]:
    if not histograms:
        return

    assert step is not None, "Step must be provided when histograms are present"

    stream = peekable(_stream_histograms(histograms))
    step = make_step(step)
    timestamp = datetime_to_proto(timestamp)

    # Local bindings for faster name lookups
    _proto_string_size = proto_string_size
    _peek_stream = stream.peek
    while stream:
        update = UpdateRunSnapshot(step=step, timestamp=timestamp)

        size = 0
        while size < max_size:
            try:
                key, histogram = _peek_stream()
            except StopIteration:
                break

            new_size = size + _proto_string_size(key) + histogram.ByteSize() + 6
            if new_size > max_size:
                break

            update.append[key].histogram.CopyFrom(histogram)
            size, _ = new_size, next(stream)

        yield update


def _stream_histograms(histograms: dict[str, Histogram]) -> Iterator[tuple[str, ProtobufHistogram]]:
    # local bindings, faster in tight loops
    _is_instance = isinstance
    _max_histogram_bin_edges = MAX_HISTOGRAM_BIN_EDGES

    for key, value in _validate_paths(histograms):
        if not _is_instance(value, Histogram):
            _warn_or_raise_on_invalid_value(f"Histogram values must be of type Histogram (got `{key}`:`{value}`)")
            continue

        has_counts, has_densities = value.counts is not None, value.densities is not None
        values_field = value.counts if has_counts else value.densities

        try:
            bin_edges = value.bin_edges_as_list()
            counts = value.counts_as_list() if value.counts is not None else None
            densities = value.densities_as_list() if value.densities is not None else None
        except TypeError as e:
            _warn_or_raise_on_invalid_value(f"{e} (at `{key}`)")
            continue

        # Merge the check for 'either one but not both is set' into a single condition
        if has_counts == has_densities:
            _warn_or_raise_on_invalid_value(
                f"One of Histogram counts and densities must be set, and they cannot be set together (at `{key}`)"
            )
            continue

        if not bin_edges or len(bin_edges) > _max_histogram_bin_edges:
            _warn_or_raise_on_invalid_value(
                f"Histogram bin edges must not be empty and must have at most {_max_histogram_bin_edges} elements "
                f"(got {len(value.bin_edges)} bin edges at `{key}`)"
            )
            continue

        if len(values_field) != len(value.bin_edges) - 1:  # type: ignore
            field_name = "counts" if has_counts else "densities"
            _warn_or_raise_on_invalid_value(
                f"Histogram {field_name} must be of length equal to bin_edges - 1 "
                f"(got {len(values_field)} {field_name} and {len(value.bin_edges)} bin edges at `{key}`)"  # type: ignore
            )
            continue

        try:
            histogram = ProtobufHistogram(
                bin_edges=bin_edges,
                counts=ProtobufHistogram.Counts(values=counts) if has_counts else None,
                densities=ProtobufHistogram.Densities(values=densities) if has_densities else None,
            )
        # TypeError is raised by protobuf when the values are not numeric, also if counts are not int
        except TypeError:
            _warn_or_raise_on_invalid_value(
                f"Histogram bin_edges and densities must be numeric, counts must be int, if provided (at `{key}`)"
            )
            continue

        yield key, histogram


def _is_over_utf8_bytes_limit(string: str, max_bytes: int) -> bool:
    """Return True if a given string can NOT fit into `max_bytes` when encoded as UTF-8"""

    # Don't encode the string if it is shorter than this value.
    # A UTF8 encoded character can take up to 4 bytes, so if we know the string will fit,
    # there's no need to encode to check length.
    encode_length_threshold = max_bytes // 4

    return len(string) > encode_length_threshold and len(string.encode("utf-8")) > max_bytes


def datetime_to_proto(dt: datetime) -> Timestamp:
    dt_ts = dt.timestamp()
    return Timestamp(seconds=int(dt_ts), nanos=int((dt_ts % 1) * 1e9))


def decompose_step(step: Union[float, int]) -> tuple[int, int]:
    """Decompose a number representing a step into whole and micro parts."""

    m = 1e6
    micro_total = int(round(step * m, 6))
    whole = int(micro_total // m)
    micro = int(micro_total % m)

    return whole, micro


def make_step(number: Union[float, int]) -> Step:
    """
    Converts a number to protobuf Step value. Example:
    >>> assert make_step(7.654321) == Step(whole=7, micro=654321)

    Args:
        number: step expressed as number

    Returns: Step protobuf used in Neptune API.
    """
    whole, micro = decompose_step(number)
    return Step(whole=whole, micro=micro)


def _warn_or_raise_on_invalid_value(message: str) -> None:
    if INVALID_VALUE_ACTION == "drop":
        logger.warning(f"Dropping value. {message}.")
    else:
        raise NeptuneUnableToLogData(message)



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

__all__ = ("ApiClient", "with_api_errors_handling")

import functools
from collections.abc import (
    Callable,
    Iterable,
    Mapping,
)
from dataclasses import dataclass
from json import JSONDecodeError
from typing import (
    Any,
    Literal,
)

import httpx
from google.protobuf.message import DecodeError as ProtobufDecodeError
from httpx import Timeout
from neptune_api import (
    AuthenticatedClient,
    Client,
)
from neptune_api.api.backend import get_client_config
from neptune_api.api.ingestion import (
    bulk_check_status,
    ingest,
)
from neptune_api.api.storage import (
    complete_multipart_upload,
    signed_url_generic,
)
from neptune_api.auth_helpers import exchange_api_key
from neptune_api.credentials import Credentials
from neptune_api.errors import (
    ApiKeyRejectedError,
    InvalidApiTokenException,
    UnableToDeserializeApiKeyError,
    UnableToExchangeApiKeyError,
    UnableToParseResponse,
    UnableToRefreshTokenError,
)
from neptune_api.models import (
    ClientConfig,
    CompleteMultipartUploadRequest,
    CreateSignedUrlsRequest,
    CreateSignedUrlsResponse,
    FileToSign,
    MultipartPart,
    Permission,
)
from neptune_api.proto.neptune_pb.ingest.v1.pub.client_pb2 import (
    RequestId,
    RequestIdList,
)
from neptune_api.proto.neptune_pb.ingest.v1.pub.ingest_pb2 import RunOperation
from neptune_api.types import File as BinaryContent
from neptune_api.types import Response

from neptune_scale.exceptions import (
    NeptuneConnectionLostError,
    NeptuneInvalidCredentialsError,
    NeptuneUnableToAuthenticateError,
    NeptuneUnexpectedResponseError,
)
from neptune_scale.sync.parameters import HTTP_CLIENT_NETWORKING_TIMEOUT
from neptune_scale.util.envs import (
    ALLOW_SELF_SIGNED_CERTIFICATE,
    VERIFY_SSL,
    get_bool,
)
from neptune_scale.util.logger import get_logger

logger = get_logger()


@dataclass
class TokenRefreshingURLs:
    authorization_endpoint: str
    token_endpoint: str

    @classmethod
    def from_dict(cls, data: dict) -> TokenRefreshingURLs:
        return TokenRefreshingURLs(
            authorization_endpoint=data["authorization_endpoint"], token_endpoint=data["token_endpoint"]
        )


@dataclass
class FileSignRequest:
    path: str
    size: int
    permission: Literal["read", "write"]


def get_config_and_token_urls(
    *,
    credentials: Credentials,
    verify_ssl: bool,
) -> tuple[ClientConfig, TokenRefreshingURLs]:
    with Client(
        base_url=credentials.base_url,
        follow_redirects=True,
        verify_ssl=verify_ssl,
        timeout=Timeout(timeout=HTTP_CLIENT_NETWORKING_TIMEOUT),
        headers={"User-Agent": _generate_user_agent()},
    ) as client:
        try:
            config_response = get_client_config.sync_detailed(client=client)
            if config_response.status_code != 200:
                raise NeptuneUnexpectedResponseError()

            config = config_response.parsed
            urls_response = client.get_httpx_client().get(config.security.open_id_discovery)
            if not urls_response.is_success:
                raise NeptuneUnexpectedResponseError()

            token_urls = TokenRefreshingURLs.from_dict(urls_response.json())

            return config, token_urls
        except JSONDecodeError:
            raise NeptuneUnexpectedResponseError()


def create_auth_api_client(
    *,
    credentials: Credentials,
    config: ClientConfig,
    token_refreshing_urls: TokenRefreshingURLs,
    verify_ssl: bool,
) -> AuthenticatedClient:
    return AuthenticatedClient(
        base_url=credentials.base_url,
        credentials=credentials,
        client_id=config.security.client_id,
        token_refreshing_endpoint=token_refreshing_urls.token_endpoint,
        api_key_exchange_callback=exchange_api_key,
        follow_redirects=True,
        verify_ssl=verify_ssl,
        timeout=Timeout(timeout=HTTP_CLIENT_NETWORKING_TIMEOUT),
        headers={"User-Agent": _generate_user_agent()},
    )


_ILLEGAL_CHARS = str.maketrans({c: "_" for c in " ();/"})


def _generate_user_agent() -> str:
    import platform
    from importlib.metadata import version

    def sanitize(value: Callable[[], str]) -> str:
        try:
            result = value()
            return result.translate(_ILLEGAL_CHARS)
        except Exception:
            return "unknown"

    package_name = "neptune-scale"
    package_version = sanitize(lambda: version(package_name))
    additional_metadata = {
        "neptune-api": sanitize(lambda: version("neptune-api")),
        "python": sanitize(platform.python_version),
        "os": sanitize(platform.system),
    }

    additional_metadata_str = "; ".join(f"{k}={v}" for k, v in additional_metadata.items())
    return f"{package_name}/{package_version} ({additional_metadata_str})"


class ApiClient:
    def __init__(self, api_token: str) -> None:
        credentials = Credentials.from_api_key(api_key=api_token)

        verify_ssl: bool = get_bool(
            VERIFY_SSL,
            default_invalid=True,
            default_missing=not get_bool(ALLOW_SELF_SIGNED_CERTIFICATE, default_missing=False, default_invalid=False),
        )

        logger.debug("Trying to connect to Neptune API")
        config, token_urls = get_config_and_token_urls(credentials=credentials, verify_ssl=verify_ssl)
        self.backend = create_auth_api_client(
            credentials=credentials,
            config=config,
            token_refreshing_urls=token_urls,
            verify_ssl=verify_ssl,
        )
        logger.debug("Connected to Neptune API")

    def submit(self, operation: RunOperation, family: str) -> Response[BinaryContent]:
        return ingest.sync_detailed(client=self.backend, body=BinaryContent(payload=operation.SerializeToString()))

    def check_batch(self, request_ids: list[str], project: str) -> Response[BinaryContent]:
        return bulk_check_status.sync_detailed(
            client=self.backend,
            project_identifier=project,
            body=BinaryContent(
                payload=RequestIdList(
                    ids=[RequestId(value=request_id) for request_id in request_ids]
                ).SerializeToString()
            ),
        )

    def close(self) -> None:
        logger.debug("Closing API client")
        self.backend.__exit__()

    def fetch_file_storage_urls(
        self,
        file_sign_requests: Iterable[FileSignRequest],
        project: str,
    ) -> Response[CreateSignedUrlsResponse]:
        files_to_sign = [
            FileToSign(
                path=request.path,
                project_identifier=project,
                permission=Permission(request.permission),
                size=request.size,
            )
            for request in file_sign_requests
        ]
        return signed_url_generic.sync_detailed(
            client=self.backend,
            body=CreateSignedUrlsRequest(
                files=files_to_sign,
            ),
        )

    def complete_multipart_upload(
        self, upload_id: str, project: str, path: str, etags: Mapping[int, str]
    ) -> Response[Any]:
        return complete_multipart_upload.sync_detailed(
            client=self.backend,
            body=CompleteMultipartUploadRequest(
                upload_id=upload_id,
                project_identifier=project,
                path=path,
                parts=[
                    MultipartPart(
                        part_number=ix,
                        etag=etag,
                    )
                    for ix, etag in etags.items()
                ],
            ),
        )


def with_api_errors_handling(func: Callable[..., Any]) -> Callable[..., Any]:
    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return func(*args, **kwargs)
        except (InvalidApiTokenException, UnableToDeserializeApiKeyError, ApiKeyRejectedError) as e:
            raise NeptuneInvalidCredentialsError() from e
        except (UnableToRefreshTokenError, UnableToExchangeApiKeyError) as e:
            # The errors above are raised by neptune-api when API token retrieval/refresh fails for
            # reasons other than the token being explicitly rejected by the server: network errors,
            # HTTP status != 200 etc. We should retry on these.
            raise NeptuneUnableToAuthenticateError() from e
        except httpx.RequestError as e:
            raise NeptuneConnectionLostError() from e
        except (JSONDecodeError, ProtobufDecodeError, UnableToParseResponse) as e:
            raise NeptuneUnexpectedResponseError() from e
        except Exception as e:
            raise e

    return wrapper
